/* flags.hh
   edited by mater 060705
   added comments for readability */

#ifndef __FLAGS_HH__
#define __FLAGS_HH__

class Flags {

public:
  Flags();
  ~Flags();

  static void usage(void);

  static bool b_vcg; // Generate DAGs in VCG format. -vcg
  static bool b_dot; // Generate DAGs in dot format. -dot
  static int i_bb;   // basic block number. bb=n
  static int i_hs;   // Heuristic: at each z generate at most n variants. -hs=n
  static int i_ts;   // Heuristic: at each stage generates at most n transfers. -ts=n
  static bool b_ilp; // Generate ILP data. -ilp
  static bool no_ddep_link; // Remove DDEP and LINK edges (resulting code might be erronous)\n -ddep

  typedef enum { TIME, ENERGY, LST, GA } opt_type_t;
  static opt_type_t opt_type; // Optimize for time(t) or energy(e)};

  // Genetic algorithm flags
  static bool b_ga_pthreads; // Parallel creation of schedules (default false)
  static unsigned int i_ga_max_gen; // Maximum number of generations to run.
  static unsigned int i_ga_size; // Number of individuals to use.
  static unsigned int i_ga_stable_stop; // Stop evolution after this number of generations with stable fitness.
  static double d_ga_mutate_node; // Percentage of node mutations.
  static double d_ga_mutate_inst; // Percentage of instruction priority mutations.
  static double d_ga_mutate_transfer; // Percentage of transfer instruction priority mutations.
  static double d_ga_new_children; // Percentage of new children relative to old population size.
  static bool b_ga_allow_multiple_mutations; // If we want to allow multiple mutations per child.
  static bool b_ga_parents_survive; // If the parents should survive and compete with the new childs.
  static bool b_ga_survivor_truncation; // If the survivors should be selected with truncation selection.
  static bool b_ga_survivor_roulette; // If the survivors should be selected with roulette wheel selection.
  static bool b_ga_bench; // If the output should be easier to parse; i.e. fitted for scripted benchmarking.
  static bool b_ga_random_search; // Random search instead of evolution.

  /* Added stuff for bookkeeping */
  static char* architecture_file;
  static char* program_file;
};

#endif //__FLAGS_HH__

