#ifndef __GAGNODE_HH__
#define __GAGNODE_HH__

#include <boost/graph/graph_traits.hpp>

#include "pudag.hh"
#include "tinst.hh"
#include "tp.hh"
#include "sp.hh"

#include "gainstcand.hh"


namespace GA
{


/**
 * A gene node.
 */
class GNode
{
  
public:
#ifdef _SP_
  GNode() : z_choice(boost::graph_traits<DAG>::null_vertex()), reference_time(-1), tau(-1),
            target_instruction(0), time_profile(0), space_profile(0), type(NO_INSTRUCTION) {};
  GNode(const vertex_set_t& _z, const vertex_id& _z_choice, int _reference_time,
        int _tau, TargetInstruction* _target_instruction, TimeProfile* _time_profile,
        SpaceProfile* _space_profile, InstructionType _type)
    : z(_z), z_choice(_z_choice), reference_time(_reference_time), tau(_tau),
      target_instruction(_target_instruction), time_profile(_time_profile),
      space_profile(_space_profile), type(_type) {}
#else
  GNode() : z_choice(boost::graph_traits<DAG>::null_vertex()), reference_time(-1), tau(-1),
            target_instruction(0), time_profile(0) {};
  GNode(const vertex_set_t& _z, const vertex_id& _z_choice, int _reference_time,
        int _tau, TargetInstruction* _target_instruction, TimeProfile* _time_profile,
        InstructionType _type)
    : z(_z), z_choice(_z_choice), reference_time(_reference_time), tau(_tau),
      target_instruction(_target_instruction), time_profile(_time_profile), type(_type) {}
#endif //_SP_
  GNode(const GNode& _gnode);
  ~GNode();
  const GNode& operator=(const GNode& _gnode);
  
  int get_reference_time() const { return reference_time; }
  void set_reference_time(int _reference_time) { reference_time = _reference_time; }
  
  int get_tau() const { return tau; }
  void set_tau(int _tau) { tau = _tau; }
  
  /* The zero-indegree set. */
  const vertex_set_t& get_z() const { return z; }
  vertex_set_t& get_z() { return z; }
  void set_z(const vertex_set_t& _z) { z = _z; }
  
  /* The choice we made from the current z set to get to the next node. */
  const vertex_id& get_z_choice() const { return z_choice; }
  void set_z_choice(const vertex_id& _z_choice) { z_choice = _z_choice; }
  
  TargetInstruction* get_target_instruction() const { return target_instruction; }
  void set_target_instruction(TargetInstruction* _target_instruction) {
    target_instruction = _target_instruction; }
  
  TimeProfile* get_time_profile() const { return time_profile; }
  void set_time_profile(TimeProfile* _time_profile) {
    time_profile = _time_profile; }
  
#ifdef _SP_
  SpaceProfile* get_space_profile() const { return space_profile; }
  void set_space_profile(SpaceProfile* _space_profile) {
    space_profile = _space_profile; }
#endif //_SP_
  
  InstructionType get_instruction_type() { return type; }
  void set_instruction_type(InstructionType _type) { type = _type; }
  
private:
  void copy_data(const GNode& _gnode);
  void delete_me();
  
  vertex_set_t z; // The zero-indegree set we had/have in that node
  vertex_id z_choice; // The choice we made from that zero-indegree set
  int reference_time; // Reference time, so we know when we made the choice (???)
  int tau;
  TargetInstruction* target_instruction;
  
  TimeProfile* time_profile;
#ifdef _SP_
  SpaceProfile* space_profile;
#endif //_SP_
  
  enum InstructionType type;
};

};

std::ostream& operator<<(std::ostream &, const GA::GNode&);

#endif //__GAGNODE_HH__
