#include "gaindividual.hh"

#include <iostream>

#include "gautility.hh"
#include "gagnode.hh"


namespace GA
{


Individual::Individual()
  : fitness(-1), schedule(0)
{
}

Individual::Individual(const Individual& _ind)
{
  copy_data(_ind);
}

/**
 * Destroys the Individual and deletes all its GNodes, and
 * delete the Schedule.
 */
Individual::~Individual()
{
  delete_me();
}

void Individual::delete_me()
{
  gnodes_t::iterator it = nodes.begin();
  for (; it != nodes.end(); ++it) {
    delete *it;
    *it = NULL;
  }
  nodes.clear();
  if (schedule != NULL) {
    delete schedule;
    schedule = NULL;
  }
}

const Individual& Individual::operator=(const Individual& _ind)
{
  if (&_ind != this) {
    delete_me();
    copy_data(_ind);
  }
  return *this;
}

void Individual::copy_data(const Individual& _ind)
{
  birth_date = _ind.birth_date;
  fitness  = _ind.fitness;
  z_order = _ind.z_order;
  
  if (_ind.schedule == 0) {
    schedule = 0;
  } else {
    schedule = new Schedule(*_ind.schedule);
  }
  
  instruction_priorities = _ind.instruction_priorities;
#ifdef _SP_
#ifdef _TRANS_
  transfer_priorities = _ind.transfer_priorities;
#endif
#endif
  
  gnodes_t::const_iterator it = _ind.nodes.begin();
  for (; it != _ind.nodes.end(); ++it) {
    if (*it == 0) {
      nodes.push_back(0);
    } else {
      nodes.push_back(new GNode(*(*it)));
    }
  }
}

/**
 * Calculates the fitness and returns the new fitness value.
 */
// double Individual::calculate_fitness()
// {
//   fitness = get_tau();
  
//   return fitness;
// }


/**
 * Calculates the fitness and returns the new fitness value.
 */
const fitness_t& Individual::calculate_fitness()
{
  fitness.set_value(get_tau());
  return fitness;
}


/**
 * Sets a new Schedule and returns a pointer to it. Does NOT delete
 * the old schedule.
 */
Schedule* Individual::set_schedule(Schedule* _schedule)
{
  schedule = _schedule;
  return schedule;
}

};


std::ostream & operator<<(std::ostream& _out, const GA::Individual& _ind)
{
  int num = 0;
  
  _out << "z_order: " << GA::print_zlist(_ind.get_const_z_order())
       << std::endl;
  _out << "gnodes:\n";
  GA::gnodes_t::const_iterator it = _ind.get_nodes().begin();
  for (; it != _ind.get_nodes().end(); ++it) {
    //_out << **it;
    //_out << "-----\n";
#if 1
    _out << num << ": "
         << "ref: " << (*it)->get_reference_time()
         << ", tau: " << (*it)->get_tau();

    if ((*it)->get_z_choice() != boost::graph_traits<DAG>::null_vertex()) {
      _out << ", sel: " << (*it)->get_z_choice();
    } else {
      _out << ", sel: -";
    }
    
    _out << ", z: " << GA::print_zset((*it)->get_z());
    
    
    if ((*it)->get_target_instruction()) {
      _out << " " << (*it)->get_target_instruction()->get_mnemonic();
    }
    _out << std::endl;
        
    ++num;
#endif
  }
  
  return _out;
}
