#ifndef __GAINDIVIDUAL_HH__
#define __GAINDIVIDUAL_HH__

#include <list>
#include <iostream>
#include <functional>
#include <boost/graph/graph_traits.hpp>

#include "sched.hh"
#include "pudag.hh"
#include "tinst.hh"

#include "gafitness.hh"
#include "gagnode.hh"


namespace GA
{


struct less_is_better_fitness
{
  /**
   * Returns true iff t1 has a better fitness value than t2.
   */
  template<typename T>
  static bool is_less_than(const T& t1, const T& t2)
  {
    return t1 < t2;
  };
};


typedef Fitness<double, less_is_better_fitness> fitness_t;

typedef std::vector<GNode*> gnodes_t;
typedef std::list<vertex_id> zorder_t;

struct ltvertex_id
{
  bool operator()(const vertex_id& v1, const vertex_id& v2) const
  {
    return v1 < v2;
  }
};

typedef std::map<vertex_id, TargetInstruction*, ltvertex_id> instruction_priority_t;

// TODO: Explain
/* The vector contains an entry for all transfers (even those that
   doesn't have a priority!) and if the value is != 0, it has a
   priority. If it's 0, it doesn't have a priority, but there is/was a
   transfer there.
 */
#ifdef _SP_
#ifdef _TRANS_
typedef std::map<vertex_id, std::vector<TargetInstruction*>, ltvertex_id> transfer_priority_t;
#endif
#endif



/**
 * A class for representing an individual.
 */
class Individual
{
  
public:
  Individual();
  Individual(const Individual& _ind);
  ~Individual();
  const Individual& operator=(const Individual& _ind);
  
  
  /* The fitness is the execution time. This is assumed elsewhere too,
     or at least that lower is better. An improvement would perhap be
     to make a fitness class with has operators for <,>,== and so on. */
//   double calculate_fitness();
//   double get_fitness() const { return fitness; } // Lower is better.
  
  const fitness_t& calculate_fitness();
  const fitness_t& get_fitness() const { return fitness; }
  
  // Tau in the last node (the execution time of the entire Schedule)
  int get_tau() const { return nodes.back()->get_tau(); }
  
  int get_birth_date() const { return birth_date; }
  void set_birth_date(int _birth_date) { birth_date = _birth_date; }
  
  gnodes_t& get_nodes() { return nodes; }
  const gnodes_t& get_nodes() const { return nodes; }
  zorder_t& get_z_order() { return z_order; }
  const zorder_t& get_const_z_order() const { return z_order; }
  
  instruction_priority_t& get_instruction_priorities() { return instruction_priorities; }
#ifdef _SP_
#ifdef _TRANS_
  transfer_priority_t& get_transfer_priorities() { return transfer_priorities; }
#endif
#endif
  Schedule* get_schedule() { return schedule; }
  Schedule* set_schedule(Schedule* _schedule);
  
private:
  void copy_data(const Individual& _ind);
  void delete_me();
  
  int birth_date;
  //double fitness;
  fitness_t fitness;
  gnodes_t nodes;
  Schedule* schedule;
  zorder_t z_order;
  instruction_priority_t instruction_priorities;
#ifdef _SP_
#ifdef _TRANS_
  transfer_priority_t transfer_priorities;
#endif
#endif
  
};


/**
 * A Strict Weak Ordering model that sorts in order of descending fitness.
 */
struct better_individual_first : public std::binary_function<Individual*, Individual*, bool>
{
  bool operator()(Individual* ind1, Individual* ind2) {
    return ind2->get_fitness() > ind1->get_fitness(); }
};


};


std::ostream& operator<<(std::ostream &, const GA::Individual&);


#endif //__GAINDIVIDUAL_HH__
