/* TODO: Make it possible to specifiy some parameters. */

/* Order of includes is important! */
#include "gapopulation.hh"
#include "gaindividual.hh"
#include "gautility.hh"

#include "gaopt.hh"
#include <iostream>
#include <deque>
#include "pudag.hh"
#include "emit.hh"
#include "verbose.hh"
#include <iomanip>
#include <time.h>
#include <sys/time.h>

GAOptimist::GAOptimist(Architecture *_architecture)
{
  architecture = _architecture;
}

GAOptimist::~GAOptimist()
{
}

Schedule *GAOptimist::schedule(DAG & _dag, Flags & flags)
{
  struct timeval e,s;
  bool bench = false;
  if (flags.b_ga_bench) {
    bench = true;
  }
  
  int seed = time(NULL);
  srand(seed);
  if (!bench) {
    std::cout << "Using seed: " << seed << std::endl;
  }
  
  int num_vertices = boost::num_vertices(_dag);
  std::cout << "Number of vertices: " << num_vertices << std::endl;
  
  Emitter* emitter = new Emitter(architecture);
  gettimeofday(&s,NULL);
  GA::Population pop(flags.i_ga_size, architecture, _dag, flags);
  pop.initialize();
  pop.evolve(flags.i_ga_max_gen);
  GA::Individual* best = pop.get_best_individual();
  gettimeofday(&e,NULL);
  
  if (best != NULL) {
    if (!bench) {
      std::cout << *best << std::endl;
    }
    
    if (!bench) {
      std::cout << "Fitness: " << best->get_fitness()
                << ", birth: " << best->get_birth_date() << std::endl;
    }
  }
  
  std::cout << "Wall clock time (ms): " 
	    << std::setw(8) << ((e.tv_sec - s.tv_sec)*1000000 + e.tv_usec-s.tv_usec)/1000 
	    << std::endl;

  
  delete emitter;
  
  if (!bench) {
    return new Schedule(*best->get_schedule());
  } else {
    // Since we're not actually using the schedule, we return an empty Schedule to avoid printing.
    return new Schedule(0);
  }
}
