#ifndef __PUDAG_HH__
#define __PUDAG_HH__

#include <boost/config.hpp>
#include <boost/graph/subgraph.hpp>
#include <boost/graph/adjacency_list.hpp>
#include <boost/graph/graph_utility.hpp>
#include "c.hh"

using namespace boost;

/***
 * Properties for the a DAG node:
 * - lcc::Node (reference to original node)
 * - int time_stamp (time stamp to resolve additional dependencies)
 * Properties for DAG edges:
 * - type
 */

//DAG Vertex properties
struct lcc_node_t {
	typedef vertex_property_tag kind;
};

struct time_stamp_t {
	typedef vertex_property_tag kind;
};

typedef property< lcc_node_t, lcc::Node,
	property< time_stamp_t, int >
> vertex_property_t;

//DAG Edge property
enum EDGE_T {
	KID0, 
	KID1, 
	LINK, 
	DDEP
};

struct edge_type_t {
	typedef edge_property_tag kind;
};

typedef property< edge_type_t, enum EDGE_T > edge_property_t;

typedef adjacency_list< vecS, vecS, bidirectionalS,
	vertex_property_t, edge_property_t > DAG;

typedef graph_traits< DAG >::vertex_descriptor vertex_id;
typedef property_map< DAG, lcc_node_t > lcc_node_map;
typedef property_map< DAG, time_stamp_t > time_stamp_map;

typedef graph_traits< DAG >::edge_descriptor edge_id;
typedef property_map< DAG, edge_type_t > edge_type_map;

typedef graph_traits< DAG >::out_edge_iterator out_edge_iter;
typedef graph_traits< DAG >::in_edge_iterator in_edge_iter;

typedef adjacency_iterator_generator< DAG, 
	vertex_id, out_edge_iter>::type successor_iterator;
typedef inv_adjacency_iterator_generator< DAG,
	vertex_id, in_edge_iter>::type predecessor_iterator;

/* mater */
typedef graph_traits< DAG >::vertex_iterator DAG_vertex_iterator;
/* end */

//Access methods (should be template)
lcc_node_map::type lcc_node_property(DAG & _dag);
lcc::Node get_lcc_node(vertex_id _v, DAG & _dag);
int get_time_stamp(vertex_id _v, DAG & _dag);
edge_type_map::type edge_type_property(DAG & _dag);
enum EDGE_T get_edge_type(edge_id _e, DAG & _dag);
const char *edgetype(enum EDGE_T _et);

vertex_id add_vertex(lcc::Node _p, int time_stamp, DAG & _dag);
edge_id add_edge(vertex_id _u, vertex_id _v, enum EDGE_T _et, DAG & _dag);

//u->v, u is a predecessor of v
std::pair< predecessor_iterator, predecessor_iterator >
predecessors(vertex_id, DAG &);

//u->v, v is a successor of u
std::pair< successor_iterator, successor_iterator >
successors(vertex_id, DAG &);

typedef std::set< vertex_id > vertex_set_t;
typedef std::map< vertex_id, int > indeg_t;

#endif //__PUDAG_HH__
