#ifndef __SCHED_HH__
#define __SCHED_HH__

#include <vector>
#include "pattern.hh"
#include "tinst.hh"
#include "archi.hh"
#include "node.hh"

/***
 * A NOP pattern in the schedule slot corresponds to a NOP
 * instruciton. All slots are filled with NOPs if the current
 * instruction set is ejected.
 */
class Schedule {

public:	
  Schedule(int _omega);
  ~Schedule();
  Schedule(const Schedule &);
  //return true if insertion successful, and returns offset, that is
  //additional slots that needed to be inserted. If _offset is greater
  //than 0, then the returned offset should be 0.
  std::pair< bool, int > insert(TargetInstruction *, 
				std::map< char *, std::pair< char *, lcc::Node >, ltstr > &,
				pattern_t *, iso_map_s isomaps, 
				const vertex_set_t &_chi, int _offset = 0);
  int size(void) { return s.size(); }
  void advance(int _offset);
  bool available_slot(void);
  friend std::ostream & operator<<(std::ostream &, const Schedule &);

  /* Creates the reservation table for the complete schedule and prints it */
  void print_resource_usage(Architecture* architecture);
  void print_stuff(DAG & _dag);
  bool greedy(DAG & _dag, Architecture *architecture);
public:
  /* mater: very unreadable! slot_t should be a class. */
  typedef std::pair< TargetInstruction *, 
		     std::map< char *, std::pair< char *, lcc::Node>, ltstr> > inst_instance_t;
  typedef std::pair< pattern_t *, inst_instance_t > slot_t;
  std::vector< std::vector< slot_t > > s;
  int omega;
  /* mater */
  bool operator!=(const Schedule & _sched);
  bool operator==(const Schedule & _sched);
  /* Helper debug function */
  static void print_reserv_tab(const reserv_tab_t & t) ;

private:
  void merge_reserv_tab(reserv_tab_t & base, const reserv_tab_t & new_usage, int offset);
  bool insertable      (reserv_tab_t & base, const reserv_tab_t & new_usage, int offset);
  
  /* had slot_t been a class I wouldn't have to do ugly stuff like this */
  //  std::map< slot_t* , iso_map_s > iso_maps;
  //std::vector< std::vector < vertex_id > > roots;
  std::vector< std::vector < iso_map_s > > im;
};

std::ostream & operator<<(std::ostream &, const Schedule &);

#endif //__SCHED_HH__
