// A node in the solution space: called as SelectionNode,
// ExtendedSelectionNode, etc.

#ifndef _SNODE_HH_
#define _SNODE_HH_

#include <assert.h>
//#include "c.hh"
#include "tp.hh"
#include "nodeattr.hh"
#include "sched.hh"
//#include "grid.hh"
#include "tinst.hh"
#include "sp.hh"
#include "archi.hh"
#include "node.hh"

#include <ext/hash_set>

//Common: FIXME (see opts.hh)
// typedef std::map< pvertex_id, vertex_id > node_map_t;
// typedef struct { 
//   node_map_t iso_map;
//   node_map_t arg_map;
//   vertex_id root;
// } iso_map_s;


// Selection Node
class SNode : public Node {

public:

  SNode(NodeAttr *, TimeProfile *, Architecture *);
  SNode(const SNode &);
  ~SNode();
  SNode & operator=(const SNode &);
  bool operator==(const SNode &) const;
  Schedule *get_schedule(void) const;
  NodeAttr *get_node_attr(void) const;
  std::pair< bool, int > insert(const vertex_set_t & _chi,
				//vertex_id & _root, 
				DAG & _dag,
				//node_map_t & _arg_map,
				iso_map_s & _arg_map,
				pattern_t & _pattern,
				TargetInstruction *,
				Architecture *);
  int get_reference_time(void) const;
  int get_tau(void) const;
  void set_node_attr(NodeAttr *);
  void set_reference_time(int);
  TimeProfile *get_time_profile(void) const { return time_profile; }
  size_t get_hash(void) const { return time_profile->get_hash(); }
  SpaceProfile *get_space_profile(void) { 
#ifndef _SP_
    assert(false);
    return (SpaceProfile *)0;
#else
    return space_profile; 
#endif //_SP_
  }
  std::pair< bool, std::set< int > >
  get_residence_class(vertex_id _v) { 
#ifndef _SP_
    assert(false);
    return make_pair(false, std::set<int>());
#else
    if(space_profile->find(_v) == space_profile->end()) {
      return make_pair(false, std::set<int>());
    }
    return make_pair(true, (*space_profile)[_v]);
#endif //_SP_
  }
  bool greedy() const {return is_greedy;}
  void increase_merged(unsigned int _add) {merged+=_add;}
  unsigned int get_merged() const {return merged;}
  friend std::ostream & operator<<(std::ostream &, const SNode &);
  
 private:
  NodeAttr *node_attr;
  int reference_time;    //tau: complition time, t: time profile reference time
  TimeProfile *time_profile;
#ifdef _SP_
  SpaceProfile *space_profile;
#else
  std::map<vertex_id, int> available_map;
#endif //_SP_

  Schedule *s;
  bool is_greedy;
  unsigned int merged;
};

std::ostream & operator<<(std::ostream &, const SNode &);


#endif //_SNODE_HH_
