
#include <assert.h>
#include "tinst.hh"

char *TargetInstruction::dest1 = "dest1";
char *TargetInstruction::src1 = "src1";
char *TargetInstruction::src2 = "src2";

TargetInstruction::TargetInstruction(char * _mnemonic, 
				     reserv_tab_t * _reserv_tab, 
				     int _dest1, int _src1, int _src2, 
				     char *_format, float _base_cost, std::vector< int > & _fus,
				     char *_constraint)
  : mnemonic(_mnemonic), reserv_tab(_reserv_tab), format(_format),
    base_cost(_base_cost), dest1_is(_constraint)
{

  ops_map[dest1] = _dest1;
  ops_map[src1] = _src1;
  ops_map[src2] = _src2;

  fus = _fus;

  //debug
//   std::cout << "Const(" << (void *)this << "): " << mnemonic << std::endl;
//   for(std::map< char *, int, ltstr >::iterator i = ops_map.begin();
//       i != ops_map.end(); ++i) {
//     std::cout << "op: " << i->first
// 	      << " id: " << i->second << std::endl;
//   }
  //debug
  cond_index = -1;
}

TargetInstruction::TargetInstruction(char * _mnemonic, 
				     reserv_tab_t * _reserv_tab, 
				     std::map< char *, int, ltstr > & _id_rclass_map,
				     char *_format, float _base_cost, std::vector< int > & _fus,
				     char *_constraint)
  : mnemonic(_mnemonic), reserv_tab(_reserv_tab), ops_map(_id_rclass_map), 
    format(_format), base_cost(_base_cost), dest1_is(_constraint)
{
  fus = _fus;

  //debug
//   std::cout << "Build(" << (void*)this << "): " << mnemonic << std::endl;
//   for(std::map< char *, int, ltstr >::iterator i = _id_rclass_map.begin();
//       i != _id_rclass_map.end(); ++i) {
//     std::cout << "op: " << i->first
// 	      << " id: " << i->second << std::endl;
//     ops_map[i->first] = i->second;
//   }

  cond_index = -1;

}

TargetInstruction::TargetInstruction(const TargetInstruction & _t_instr)
  : mnemonic(_t_instr.mnemonic), reserv_tab(_t_instr.reserv_tab),
    ops_map(_t_instr.ops_map), format(_t_instr.format), 
    base_cost(_t_instr.base_cost), fus(_t_instr.fus), 
    cond_index(_t_instr.cond_index), dest1_is(_t_instr.dest1_is)
{
}

TargetInstruction::~TargetInstruction()
{
}

int TargetInstruction::get_residence_class(char *_rc_name)
{ 
  //debug
//   std::cout << "Get(" << (void *)this << "): " << mnemonic << std::endl;
//   for(std::map< char *, int, ltstr >::iterator i = ops_map.begin();
//       i != ops_map.end(); ++i) {
//     std::cout << "op: " << i->first
// 	      << " id: " << i->second << std::endl;
//   }
  //debug

  assert(ops_map.find(_rc_name) != ops_map.end());
  return ops_map[_rc_name];
}

float TargetInstruction::get_base_cost(void)
{
  return base_cost;
}

std::vector< int > TargetInstruction::get_use_funits(void)
{
  return fus;
}

std::ostream & operator<<(std::ostream & _out, const TargetInstruction & _target_instruction)
{
  _out << (void *)&_target_instruction << " ";
  _out << "mnemonic: " << _target_instruction.mnemonic << " format: " << _target_instruction.format << std::endl;
  _out << "ops_map: ";
  for(std::map< char *, int, ltstr >::const_iterator iter = _target_instruction.ops_map.begin();
      iter != _target_instruction.ops_map.end(); ++iter) {
    _out << iter->first << " -> " << iter->second << " ";
  }
  _out << "base_cost: " << _target_instruction.base_cost << " ";
  _out << "cond_index: " << _target_instruction.cond_index << std::endl;
  _out << std::endl;
  return _out;
}


