/* Copyright Microsoft Corporation. All rights reserved. */

int ARM_ASSEMBLER_MODE = 0; /* declared in emit.cc */
int OPT_C_DEBUG = 0;
#include "c.h"

#define I(f) opt_##f
static char rcsid[] = "$Id: opt.c,v 1.5 2004/11/26 12:30:59 x04anded Exp $";

static int maxlocals;		/* maximum amount of local space used, in bytes */

extern Node cg_gen(Node);
extern void cg_emit(Node, char *);
extern void cg_function(Symbol, Symbol*, Symbol*, int);
extern void cg_initialize(int, char **);
extern void cg_finalize(void);


static void I(segment)(int n) { 
  if(OPT_C_DEBUG)
    print("@segment %d\n",n);
  if(ARM_ASSEMBLER_MODE)
  {
    switch(n)
    {
	case CODE:
	  print(".text\n");
	  break;
	case BSS:
	  print(".bss\n");
	break;
	case DATA:
	  print(".data\n");
	  break;
	case LIT: /* possibly .text as well */
	  print(".data\n");
	  break;
	default:
	  print("UNKNOWN SEGMENT SWITCH\n");
    }
  }
}
static void I(blockbeg)(Env *e) { 
  if(OPT_C_DEBUG)
    print("@blockbeg\n"); 
}
static void I(blockend)(Env *e) {
  if(OPT_C_DEBUG)
    print("@blockend\n"); 
}

/* address - initialize q for addressing expression p+n */
static void I(address)(Symbol q, Symbol p, long n) { 
  if(OPT_C_DEBUG)
    print("@address\n"); 
}

static void I(defaddress)(Symbol p) { 
  if(OPT_C_DEBUG)
    print("@defaddress\n");
}

static void I(defconst)(int suffix, int size, Value v) {
  if(OPT_C_DEBUG)
    print("@defconst\n");
  if(ARM_ASSEMBLER_MODE)
  {
    switch(suffix)
    {
       case I:
	 print("\t.word\t%d\n",v.i); /* only for integers*/
	 break;
	case P:
	  print("\t.word\t$_%s\n",v.p);
	  break;
	default:
	  print("\t.some constant.....?\n");
   }
  }
}

static void I(defstring)(int len, char *str) { 
  if(OPT_C_DEBUG)
    print("@defstring %s\n",str);
  if(ARM_ASSEMBLER_MODE)
  {
    int i;
    print("\t.ascii\t\"");
    for(i=0;i<len-1;i++)
    {
      if(str[i]=='\n')
	print("\\n");
      else
	print("%c",str[i]);
    }
    print("\\000\"\n");
  }
}

static void I(space)(int n) { 
  if(OPT_C_DEBUG)
    print("@space %d\n",n);
  if(ARM_ASSEMBLER_MODE) 
    print("\t.space %d\n",n);
}

static void I(defsymbol)(Symbol p) { 
  char *uname="";
  if(OPT_C_DEBUG)
    print("@defsymbol %s\n",p->name);
  
  if(ARM_ASSEMBLER_MODE)
  {
    if (p->scope == CONSTANTS )
      p->x.name = p->name;
    else if (p->sclass == STATIC && p->generated)
      p->x.name = stringf("$%s_%s", uname, p->name);
    else if (p->sclass == STATIC)
      p->x.name = stringf("$%s%d_%s", uname, genlabel(1), p->name);
    else if (p->scope == LABELS)
      p->x.name = stringf("$L%s", p->name);
    else if (p->generated)
      p->x.name = stringf("$%s", p->name);
    else
      p->x.name = stringf("%s", p->name);
  }
  else
    p->x.name = stringf("%s", p->name);
}

static void I(global)(Symbol p) {
  if(OPT_C_DEBUG)
    print("@global %s\n",p->name);
  if(ARM_ASSEMBLER_MODE)
  {
    print(".text\n");
    
    print(".%s:\n\t.word %s\n",p->x.name,p->x.name);
    
    I(segment)(p->u.seg);
    
    print("%s:\n",p->x.name);
  }
}

static void I(emit)(Node p) {
  cg_emit(p, "archi.xml");
}

static void I(local)(Symbol p) {
  if(OPT_C_DEBUG)
    print("@local %s, %d bytes\n",p->name,p->type->size);
  
  p->x.offset = 12+p->type->size+maxlocals;
  maxlocals = roundup(maxlocals + p->type->size, 4);
  p->x.name = p->name;
}

static void I(export)(Symbol p) {
  if(OPT_C_DEBUG)
    print("@export %s\n",p->name);
  if(ARM_ASSEMBLER_MODE)
    print("\t.global %s\n",p->x.name); 
}

static void I(function)(Symbol f, Symbol caller[], Symbol callee[], int ncalls) {
  if(OPT_C_DEBUG)
    print("@function\n");
  
  cg_function(f, caller, callee, ncalls);

  maxlocals = 0;
  {
    int i;
    for (i = 0; caller[i] && callee[i]; i++) { 
      caller[i]->x.offset = i; 
      callee[i]->sclass = caller[i]->sclass; 
      
      /* set offsets */
      caller[i]->x.offset = maxlocals+16;
      callee[i]->x.offset = maxlocals+16;
      maxlocals = roundup(maxlocals+caller[i]->type->align, 4);
    }
  }
  
  /* initialize */
  gencode(caller, callee); //gencode() lives in the front-end

  if(ARM_ASSEMBLER_MODE)
  {
    print("\t.align\t%d\n",2);
    print("\t.type\t%s,function\n",f->name);
    print("%s:\n",f->name);
    
    print("\tmov ip, sp\n");
    print("\tstmfd\tsp!, {r4, r5, fp, ip, lr, pc}\n");
    print("\tsub fp, ip, #4\n");
    print("\tsub sp, sp, #%d\n",maxlocals);
    
    /* store formal to memory */
    {
      int j;
      for (j = 0; caller[j] && callee[j]; j++) { 
	  print("\tstr\tr%d, [fp, #-%d]\n",j,caller[j]->x.offset);
      }
    }
    print("\n");
  }
  /* mater: printing the schedule */
  emitcode();
  
  if(ARM_ASSEMBLER_MODE)
  {
    print("\tldmea\tfp, {r4, r5, fp, sp, pc}\n");
    {
      static int end_label_id=1;
      print(".Lfe%d:\n",end_label_id);
      print("\t.size\t%s,.Lfe%d-%s\n",f->name,end_label_id,f->name);
      end_label_id++;
    }
  }
}



static Node I(gen)(Node p) {
  if(OPT_C_DEBUG)
    print("@gen\n");
  return cg_gen(p);
}

static void I(import)(Symbol p) { 
  if(OPT_C_DEBUG)
    print("@import\n");
}

static void I(progbeg)(int argc, char *argv[]) {
  cg_initialize(argc, argv); //OPT_C_DEBUG is set (etc.)
  if(OPT_C_DEBUG)
    print("@progbeg\n");
}

static void I(progend)(void) {
  if(OPT_C_DEBUG)
    print("@progend\n");
  
  cg_finalize();
}

Interface optIR = {
	1, 1, 0,	/* char */
	2, 2, 0,	/* short */
	4, 4, 0,	/* int */
	4, 4, 0,	/* long */
	4, 4, 0,	/* long long */
	4, 4, 0,	/* float */
	8, 4, 0,	/* double */
	8, 4, 0,	/* long double */
	4, 4, 0,	/* T * */
	0, 1, 0,	/* struct; so that ARGB keeps stack aligned */
	1,		/* little_endian */
	0,		/* mulops_calls */
	0,		/* wants_callb */
	0,		/* wants_argb */
	1,		/* left_to_right */
	1,		/* wants_dag */
	0,		/* unsigned_char */
	0,		/* I(address) */
	I(blockbeg),
	I(blockend),
	I(defaddress),
	I(defconst),
	I(defstring),
	I(defsymbol),
	I(emit),
	I(export),
	I(function),
	I(gen),
	I(global),
	I(import),
	I(local),
	I(progbeg),
	I(progend),
	I(segment),
	I(space),
	0,		/* I(stabblock) */
	0,		/* I(stabend) */
	0,		/* I(stabfend) */
	0,		/* I(stabinit) */
	0,		/* I(stabline) */
	0,		/* I(stabsym) */
	0,		/* I(stabtype) */
};

#undef I
