#include <iostream>
#include <string>
#include <sstream>
#include <stdlib.h> //strtof function
#include <dlfcn.h> //dlopen, dlerror
#include "archi.hh"
#include <boost/graph/graphviz.hpp>
#include "verbose.hh"

#define max(a,b) (((a)>(b))?(a):(b))

/* This code has many memory leaks. XMLString::transcode results must
   be freed with XMLString::release. */

/* Outputing patterns */
template < class T >
class pnode_writer {

public:
  pnode_writer(T _t) : t(_t) {}
  void operator()(std::ostream & _out, pvertex_id _v_id) const {
    vertex_info_map::type vim = vertex_info_property(t);
    vertex_kind_map::type vkm = vertex_kind_property(t);
    _out << "[label=\"id:\\\"" << vim[_v_id].id << "\\\"\\n";
    for(std::list< int >::iterator oi = vim[_v_id].ops.begin(); 
	oi != vim[_v_id].ops.end(); ++oi) {
      _out << *oi << "\\n";
    }
    _out << "deg: " << vim[_v_id].indeg << "\\n"
	 << "t: ";
    if(vkm[_v_id] == ARGNODE) {
      _out << "ARGNODE";
    } else {
      _out << "OPNODE";
    }
    _out << "\"]";
  }
private:

  T t;
};

template < class T >
class pedge_writer {

public:
  pedge_writer(T _t) : t(_t) {}
  void operator()(std::ostream & _out, pedge_id _e_id) const {
    pedge_type_map::type etm = edge_type_property(t);
    _out << "[label=\"";
    _out << edgetype(etm[_e_id])
	 << "\"]";
  }
  
private:
  T t;
};

void ErrorReport::warning(const SAXParseException & )
{
  //Ignore all warnings
  std::cerr << "Warning...\n";
}

void ErrorReport::error(const SAXParseException & toCatch)
{
  //fSawErrors = true;
  char *msg1 = XMLString::transcode(toCatch.getSystemId()),
    *msg2 = XMLString::transcode(toCatch.getMessage());

  std::cerr << "Error at file \"" << msg1
	    << "\", line " << (toCatch.getLineNumber())
	    << ", column " << (toCatch.getColumnNumber())
	    << "\n   Message: " << msg2 << std::endl;
  XMLString::release(&msg1);
  XMLString::release(&msg2);
}

void ErrorReport::fatalError(const SAXParseException & toCatch)
{
  //fSawErrors = true;
  char *msg1 = XMLString::transcode(toCatch.getSystemId()),
    *msg2 = XMLString::transcode(toCatch.getMessage());

  std::cerr << "Fatal Error at file \"" << msg1
	    << "\", line " << toCatch.getLineNumber()
	    << ", column " << toCatch.getColumnNumber()
	    << "\n   Message: " << msg2 << std::endl;
  XMLString::release(&msg1);
  XMLString::release(&msg2);
}

Architecture::Architecture(char *_xadml_file_name)
  : xadml_file_name(_xadml_file_name)
{
  cond_index = 0;

  //Operators to be implemented in the backend (from opt.c)

  //FIXME
  //Checks for validity of the file
  //debug
  /*
    int i = rand() % 1000;
    char file_name[25];
    sprintf(file_name, "/tmp/cond_%d.cc", i);
    std::ofstream f(file_name);
    f << "BEGIN" ;
  */
#define gop(x,n)
/* mater: #s makes it a string "s" */
#define op(x,t,s) addop(lcc::x,lcc::t,#s,#x #t);
#include "ops.h"
#undef gop
#undef op

  /* useless? */
  std::map< char *, int, ltstr >::const_iterator it = ops_map.begin();
  while(ops_map.end() != it) {
    it++;
  }
}

void Architecture::addop(int op, int type, const char *sz, const char *opname)
{

  static char list[] = { 'c', 's', 'i', 'l', 'h', 'f', 'd', 'x', 'p', 0 };
  static int sizes[] = {  
    lcc::IR->charmetric.size,
    lcc::IR->shortmetric.size,
    lcc::IR->intmetric.size,
    lcc::IR->longmetric.size,
    lcc::IR->longlongmetric.size,
    lcc::IR->floatmetric.size,
    lcc::IR->doublemetric.size,
    lcc::IR->longdoublemetric.size,
    lcc::IR->ptrmetric.size
  };

  static int last;
  char opname_s[10];

  if (op == lcc::LOAD)
    return;
  
  last = op;
  if (type == lcc::B || type == lcc::V) {
    ops_map[strdup(opname)] = op + type;
  } else {
    int i, done = 0;
    const char *s;
    for (i = 0; sz[i] != '\0' && (s = strchr(list, sz[i])) != NULL; i++) {
      int n = sizes[s-list];
      if ((done&(1<<n)) == 0) {
	sprintf(opname_s, "%s%d", opname, n);
	ops_map[strdup(opname_s)] = op + type + sizeop(n);
      }
      done |= 1<<n;
    }
  }
}

void Architecture::parse(void)
{
  try
    {
      XMLPlatformUtils::Initialize();
    }
  catch (const XMLException & toCatch)
    {
      char *message = XMLString::transcode(toCatch.getMessage ());
      std::cerr << "Parser:Error during initialization! :\n" << message << "\n";
      XMLString::release (&message);
      throw (void *)NULL;
    }

  parser = new XercesDOMParser();
  parser->setValidationScheme (XercesDOMParser::Val_Never); // optional. (DTD)
  //  parser->setDoNamespaces (false);	// optional

  ErrorHandler *errHandler = (ErrorHandler *) new ErrorReport();
  parser->setErrorHandler (errHandler);

  try
    {
      parser->parse(xadml_file_name);
    }
  catch(const XMLException & toCatch)
    {
      char *message = XMLString::transcode (toCatch.getMessage ());
      std::cerr << "Exception message is: \n" << message << "\n";
      XMLString::release (&message);
      throw (void *)NULL;
    }
  catch(const DOMException & toCatch)
    {
      char *message = XMLString::transcode (toCatch.msg);
      std::cerr << "Exception message is: \n" << message << "\n";
      XMLString::release (&message);
      throw (void *)NULL;
    }
  catch(...)
    {
      std::cerr << "Unexpected Exception \n";
      throw (void *)NULL;
    }

  spec = parser->getDocument();

  delete errHandler;

  //Initialize
  initialize();
}

Architecture::~Architecture()
{
  delete parser;
  op_map_t::iterator oi = op_map.begin();
  for(; oi != op_map.end(); ++oi){
    std::vector< pattern_instr_t * > & cur = (*oi).second;
    while(! cur.empty() ){
      //delete cur.back()->first; //why not this one?
      delete cur.back()->second;
      delete cur.back();
      cur.pop_back();
    }
  }
  std::map< char *, int, ltstr >::iterator it;
  it=ops_map.begin();
  for(; it!=ops_map.end(); ++it){
    free( (*it).first); //the string is strdup():ed
  }

  XMLPlatformUtils::Terminate();
  /* File only open if there actually are conditions */
  if(cond_index > 0){
    dlclose(handle);
  }
}

DOMNode *Architecture::get_sub_tree(const char *_node_name, DOMNode *_current_node)
{
  XMLCh * tmp;
  DOMNode *pCursor;
  if(!_current_node) {
    DOMElement* root = spec->getDocumentElement();
    pCursor = root->getFirstChild();
  } else {
    pCursor = _current_node;
  }
  while (pCursor != NULL)
    {
      tmp = XMLString::transcode(_node_name);
      if( pCursor->getNodeType() == DOMNode::ELEMENT_NODE &&
	  !XMLString::compareString(pCursor->getNodeName(), 
				    tmp))
	{
	  if(tmp) XMLString::release(&tmp);
	  return pCursor;
	}
      if(tmp) XMLString::release(&tmp);
      pCursor = pCursor->getNextSibling();
    }
  return NULL;
}

int Architecture::get_omega(void)
{
  DOMNode *pOmega = get_sub_tree("omega");
  if(pOmega) {
    int omega = XMLString::parseInt(pOmega->getFirstChild()->getNodeValue());
    return omega;
  } 
  return 0;
}

char *Architecture::get_defined_string(char *key)
{
  DOMNode *p = get_sub_tree(key);
  if(p)
    {
      return XMLString::transcode(p->getFirstChild()->getNodeValue());
    }
  return 0;
}

std::pair< match_iter, match_iter >
Architecture::match_start(int _op)
{
  if(op_map.find(_op) == op_map.end()) {
    std::cerr << "Warning - operand: " << _op << " : " << lcc::opname(_op) << " is not specifed in the HW spec!\n";
    std::vector< pattern_instr_t *> dummy;
    return std::make_pair(dummy.begin(), dummy.end());
    //    exit(-1);
  }
  std::vector< pattern_instr_t * > & tmp = op_map[_op];
  return std::make_pair(tmp.begin(), tmp.end());
}

char *Architecture::get_reg_name(int _reg)
{
  assert(_reg < register_index.size());
  return register_index[_reg];
}

bool Architecture::remove_op(int _op)
{
  return (remove_ops.find(_op) != remove_ops.end());
}

/* Generates architecture related ilp-data (everything but nodes and
   edges). This function is called from cg.cc . */
void Architecture::ilp(std::ofstream & _f)
{
  std::vector< struct data_s > ilp_patterns;
  std::vector< struct singleton_s > singletons;
  std::vector< struct singleton_s > nonissue;

  int vertex_number_base = 0;
  XMLCh * xinstruction = XMLString::transcode("instruction");
  XMLCh * xtarget      = XMLString::transcode("target");
  XMLCh * xid          = XMLString::transcode("id");
  XMLCh * xcycle_matrix= XMLString::transcode("cycle_matrix");
  XMLCh * xlatency     = XMLString::transcode("latency");
  XMLCh * xl           = XMLString::transcode("l");

  //Singleton instructions
  DOMNode *instructionset = get_sub_tree("instructionset");
  assert(instructionset);
  DOMNode *instruction = instructionset->getFirstChild();

  while(instruction != NULL) {
    if(instruction->getNodeType() == DOMNode::ELEMENT_NODE) {
      if(!XMLString::compareString(instruction->getNodeName(), xinstruction)) {
	char *instr_id = XMLString::transcode(((DOMElement *)instruction)->getAttribute(xid));
	int op_number = get_op_number(instr_id);
	DOMNode *target = instruction->getFirstChild();
	while(target != NULL) {
	  if(target->getNodeType() == DOMNode::ELEMENT_NODE &&
	     !XMLString::compareString(target->getNodeName(), xtarget)) {
	    char *target_id = XMLString::transcode(((DOMElement *)target)->getAttribute(xid));
	    DOMNode * target_children = target->getFirstChild();
	    bool found_cycle_matrix = false;
	    while(target_children){
	      if(!XMLString::compareString(target_children->getNodeName(), xcycle_matrix)){
		found_cycle_matrix = true;
		break;
	      }
	      target_children = target_children->getNextSibling();
	    }
	    if( ! found_cycle_matrix ){
	      nonissue.push_back(singleton_s());
	      singleton_s & s = nonissue.back();
	      s.instruction_id = std::string(target_id);
	      s.op_id = op_number;
	    }

	    DOMNode *ops = target->getFirstChild();
	    int ell = 0;
	    while(ops != NULL) {
	      if(ops->getNodeType() == DOMNode::ELEMENT_NODE) {

		if(!XMLString::compareString(ops->getNodeName(), xcycle_matrix)) {
		  //Extract latency (ILP specific)
		  DOMNode *latency = ops->getFirstChild();
		  while(latency != NULL) {
		    if(!XMLString::compareString(latency->getNodeName(), xlatency)) {
		      ell = max(ell,XMLString::parseInt(((DOMElement *)latency)->getAttribute(xl)));
		    }
		    latency = latency->getNextSibling();
		  }

		  DOMNamedNodeMap *anm = ops->getAttributes();
		  bool pushed = false;
		  for(int i = 0; i < anm->getLength(); i++) {
		    DOMNode *att = anm->item(i);
		    if(att->getNodeType() != DOMNode::ATTRIBUTE_NODE) {
		      continue;
		    }
		    char *stage_name = XMLString::transcode(att->getNodeName());
		    int length = XMLString::parseInt(((DOMAttr *)att)->getValue()) + ell;
		    if(!pushed){
		      pushed = true;
		      singletons.push_back(singleton_s());
		      singleton_s & s = singletons.back();
		      s.instruction_id = std::string(target_id);
		      s.op_id = op_number;
		      s.latency = length;
		    }
		    singleton_s & s = singletons.back();
		    s.resources.insert(std::string(stage_name));
		  }
		}
	      }
	      ops = ops->getNextSibling();
	    }
	  }
	  target = target->getNextSibling();
	}
      }
    }
    instruction = instruction->getNextSibling();
  }

  //Pattern instructions
  instructionset = get_sub_tree("instructionset");
  assert(instructionset);
  DOMNode *pattern = instructionset->getFirstChild();

  /* factor is the multiplicity of every pattern in the ilp-data. It
     should be the maximum number of times the pattern can be
     used. Perhaps counting operators would be a good idea?*/
  /*** I fixed this, factor is not needed in the new cluster-ilp
       formulation. */
  int factor = 1; //size of the DAG (FIXME: to be optimized)
  std::set< int > set_pn;
  set_pn.insert(0); /* For the single node patterns */

  while(pattern != NULL) {
    if(pattern->getNodeType() == DOMNode::ELEMENT_NODE) {
      if(!XMLString::compareString(pattern->getNodeName(), XMLString::transcode("pattern"))) {
	char *pattern_id = XMLString::transcode(((DOMElement *)pattern)->getAttribute(XMLString::transcode("id")));
	DOMNode *ptarget = pattern->getFirstChild();
	int ell = 0;
	while(ptarget != NULL) {
	  if(ptarget->getNodeType() == DOMNode::ELEMENT_NODE &&
	     !XMLString::compareString(ptarget->getNodeName(), XMLString::transcode("ptarget"))) {
	    char *ptarget_id = XMLString::transcode(((DOMElement *)ptarget)->getAttribute(XMLString::transcode("id")));
	    //Name of the target instruction

	    DOMNode *ops = ptarget->getFirstChild();

	    while(ops != NULL) {
	      if(ops->getNodeType() == DOMNode::ELEMENT_NODE) {

		if(!XMLString::compareString(ops->getNodeName(), XMLString::transcode("cycle_matrix"))) {

		  //Extract latency (ILP specific) : FIXME (duplicated code, see instruction part)
		  DOMNode *latency = ops->getFirstChild();

		  while(latency != NULL) {
		    if(!XMLString::compareString(latency->getNodeName(), XMLString::transcode("latency"))) {
		      ell = max(ell, XMLString::parseInt(((DOMElement *)latency)->getAttribute(XMLString::transcode("l"))));
		    }
		    latency = latency->getNextSibling();
		  }

		  DOMNamedNodeMap *anm = ops->getAttributes();
		  for(int i = 0; i < anm->getLength(); i++) {
		    DOMNode *att = anm->item(i);
		    if(att->getNodeType() != DOMNode::ATTRIBUTE_NODE) {
		      continue;
		    }
		    char *stage_name = XMLString::transcode(att->getNodeName());
		    int length = XMLString::parseInt(((DOMAttr *)att)->getValue()) + ell;

		    char target_pattern_id[100];
		    for(int f = 0; f < factor; f++) {
		      ilp_patterns.push_back(data_s());
		      data_s & p = ilp_patterns.back();

		      //Pattern id target level
		      /* factor > 1 will cause name conflicts. setting
			 it to 1 makes PR[RC] simpler */
		      //sprintf(target_pattern_id, "%s_%d", ptarget_id, f);
		      sprintf(target_pattern_id, "%s", ptarget_id);
		      p.target_pattern_id = std::string(target_pattern_id);

		      pattern_t & pattern = *(pattern_map[pattern_id]);
		      int n_vertices = 0;// boost::num_vertices(pattern);
					 // only count real vertices!
		      pvertex_iter vi, vi_end;
		      for(tie(vi, vi_end) = boost::vertices(pattern); vi != vi_end; ++vi) {
			int v_id = (int)*vi; // removed this, we don't need it! + vertex_number_base;
			int out_deg = boost::out_degree(*vi, pattern);
			int op_id = 0;
			vertex_info_map::type vim = vertex_info_property(pattern);
			//FIXME (list of nodes for ILP will not work)
			for(std::list< int >::iterator oi = vim[*vi].ops.begin();
			    oi != vim[*vi].ops.end(); ++oi) {
			  op_id = *oi;
			  break;
			}
			/* mater: try to fix the above, only get real
			   nodes (that has operator). */
			if(op_id != 0){
			  p.nodes.push_back(std::make_pair(v_id, std::make_pair(op_id, out_deg)));
			  set_pn.insert((int)*vi);
			  n_vertices++;
			}
		      }

		      graph_traits< pattern_t >::edge_iterator ei, ei_end;
		      for(tie(ei, ei_end) = edges(pattern); ei != ei_end; ++ei) {
			int src, dest;
			vertex_info_map::type vim = vertex_info_property(pattern);
			src = (int)boost::source(*ei, pattern);// removed + vertex_number_base;
			dest = (int)boost::target(*ei, pattern); //removed + vertex_number_base;
			if(vim[boost::source(*ei,pattern)].ops.size() != 0 &&
			   vim[boost::target(*ei,pattern)].ops.size() != 0){
			  p.edges.push_back(std::make_pair(src,dest));
			}
		      }
		      vertex_number_base += n_vertices;

		      p.latency = length;
		      p.resource = (std::string(stage_name));

		    }
		  }
		}
	      }
	      ops = ops->getNextSibling();
	    }
	  }
	  ptarget = ptarget->getNextSibling();
	}
      }
    }
    pattern = pattern->getNextSibling();
  }


  std::vector< struct data_s >::iterator pi;
  std::vector< struct singleton_s>::iterator si;

  //Issue width
  _f << "# Issue width"
     << std::endl
     << "param W := " << get_omega() << ";\n" << std::endl;

#if 0 //removed for cluster
  //Number of registers
  _f << "param R := " << register_map.size() << ";\n" << std::endl;
#endif

  //Max time
  _f << "param max_t := 15; #To be adjusted\n" << std::endl;

  _f << "# Non-singleton patterns" << std::endl;
  _f << "set P_prime :=" << std::endl;
  //  std::map< std::string, std::pair< std::vector< std::pair< int, std::pair< int , int > > >, 
  //    std::vector< std::pair< int, int > > > >::iterator pi;
  for(pi = ilp_patterns.begin(); pi != ilp_patterns.end(); ++pi) {
    _f << "  " << (*pi).target_pattern_id << std::endl;
  }
  _f << ";\n" << std::endl;

  _f << "# Singleton patterns" << std::endl;
  _f << "set P_second :=" << std::endl;
  for(si = singletons.begin(); si != singletons.end(); ++si) {
    _f << "  " << (*si).instruction_id << std::endl;
  }
  _f << ";\n" << std::endl;

  _f << "# Nonissue patterns" << std::endl;
  _f << "set P_nonissue :=" << std::endl;
  for(si = nonissue.begin(); si != nonissue.end(); ++si) {
    _f << "  " << (*si).instruction_id << std::endl;
  }
  _f << ";\n" << std::endl;

  _f << "# Operator of patterns" << std::endl;
  _f << "param OPP :=" << std::endl;
  //Nonissue
  for(si = nonissue.begin(); si != nonissue.end(); ++si) {
    _f << "  " << (*si).instruction_id << "\t" << "0\t" << (*si).op_id << std::endl;
  }
  //Instructions
  for(si = singletons.begin(); si != singletons.end(); ++si) {
    _f << "  " << (*si).instruction_id << "\t" << "0\t" << (*si).op_id << std::endl;
  }

  //Patterns
  for(pi = ilp_patterns.begin(); pi != ilp_patterns.end(); ++pi) {
    std::vector< std::pair< int, std::pair< int, int > > > vertices = (*pi).nodes;
    for(int i = 0; i < vertices.size(); i++) {
      int op_id = vertices[i].second.first;
      int vertex = vertices[i].first;
      _f << "  " << (*pi).target_pattern_id << "\t" << vertex << "\t" << op_id << std::endl;
    }
  }
  _f << ";\n" << std::endl;

  //Latencies

  _f << "# Latencies of patterns" << std::endl;
  _f << "param L :=" << std::endl;
  //Instructions
  for(si = singletons.begin(); si != singletons.end(); ++si) {
    _f << "  " << (*si).instruction_id << "\t" << (*si).latency << std::endl;
  }
  //Patterns
  for(pi = ilp_patterns.begin(); pi != ilp_patterns.end(); ++pi) {
    _f << "  " << (*pi).target_pattern_id << "\t" << (*pi).latency << std::endl;
  }
  _f << ";\n" << std::endl;

  _f << "# Outdegree of (composite) patterns" << std::endl;
  _f << "param ODP :=" << std::endl;
  for(pi = ilp_patterns.begin(); pi != ilp_patterns.end(); ++pi) {
    std::vector< std::pair< int, std::pair< int, int > > > vertices = (*pi).nodes;
    for(int i = 0; i < vertices.size(); i++) {
      int out_deg = vertices[i].second.second;
      int vertex = vertices[i].first;
      if(out_deg > 0) {
	_f << "  " << (*pi).target_pattern_id << "\t" << vertex << "\t" << out_deg << std::endl;
      }
    }
  }
  _f << ";\n" << std::endl;

  //instructions
  _f << "# Instruction pattern nodes (within 0 .. n-1)" << std::endl;
  for(si = singletons.begin(); si != singletons.end(); ++si) {
    _f << "set B[" << (*si).instruction_id << "] := 0;" << std::endl;
  }
  for(si = nonissue.begin(); si != nonissue.end(); ++si) {
    _f << "set B[" << (*si).instruction_id << "] := 0;" << std::endl;
  }
  //patterns
  for(pi = ilp_patterns.begin(); pi != ilp_patterns.end(); ++pi) {
    _f << "set B[" << (*pi).target_pattern_id << "] :=";
    std::vector< std::pair< int, std::pair< int, int > > > vertices = (*pi).nodes;
    for(int i = 0; i < vertices.size(); i++) {
      _f << " " << vertices[i].first;
    }
    _f << ";" << std::endl;
  }
  _f << std::endl;

  //edges
  for(pi = ilp_patterns.begin(); pi != ilp_patterns.end(); ++pi) {
    _f << "set EP[" << (*pi).target_pattern_id << "] :=";
    std::vector< std::pair< int, int > > edges = (*pi).edges;
    for(int i = 0; i < edges.size(); i++) {
      _f << " (" << edges[i].first << "," << edges[i].second << ")";
    }
    _f << ";" << std::endl;
  }
  _f << std::endl;

  //pattern nodes
  _f << "set PN := " << std::endl;
  for(std::set< int >::iterator pni = set_pn.begin(); 
      pni != set_pn.end(); ++pni){
    _f << "  " << *pni << std::endl;
  }
  _f << ";\n\n";

  //resources
  _f << "# Resources" << std::endl;
  std::map< char *, int, ltstr >::iterator ri;
  _f << "set F :=" << std::endl;
  for(ri = stage_map.begin(); ri != stage_map.end(); ++ri) {
    _f << "  " << (*ri).first << std::endl;
  }
  _f << ";\n" << std::endl;

  //single instance of resources
  _f << "param M :=" << std::endl;
  for(ri = stage_map.begin(); ri != stage_map.end(); ++ri) {
    _f << "  " << (*ri).first << "\t1" << std::endl;
  }
  _f << ";\n" << std::endl;

  //resource usage
  _f << "# Resource usage of patterns" << std::endl;
  _f << "param U :=" << std::endl;
  //instructions
  for(si = singletons.begin(); si != singletons.end(); ++si) {
    std::set< std::string >::iterator resit;
    for(resit = (*si).resources.begin(); resit != (*si).resources.end(); ++resit){
      _f << "  " << (*si).instruction_id << "\t" << (*resit) << "\t1" << std::endl;
    }
  }
  //patterns
  for(pi = ilp_patterns.begin(); pi != ilp_patterns.end(); ++pi) {
    _f << "  " << (*pi).target_pattern_id << "\t" << (*pi).resource << "\t1" << std::endl;
  }
  _f << ";\n" << std::endl;
  
  /* mater: cluster stuff */
#define XSTR(x) XMLString::transcode(x)
  /* The set of residences, RS */
  _f << "set RS :=" << std::endl;
  std::map< char * , std::vector< int >, ltstr>::iterator residenceit;
  std::set< char * >::iterator all_res_it;
  for(all_res_it = all_residences.begin(); 
      all_res_it != all_residences.end(); 
      ++all_res_it){
    _f << "  " << (*all_res_it) << std::endl;
  }
  _f << ";" << std::endl;

  /* Patterns belonging to a residence, PR */
  for(all_res_it = all_residences.begin(); 
      all_res_it != all_residences.end(); 
      ++all_res_it){
    char *current_residence = (*all_res_it);
    std::set< char * > prdest1;
    std::set< char * > prsrc1;
    std::set< char * > prsrc2;

    /* For for every instruction */
    DOMNode *instructionset = get_sub_tree("instructionset");
    assert(instructionset);
    DOMNode *instruction = instructionset->getFirstChild();
    DOMNode *target;
    while(instruction){
      /* If dest1 represents the current RS, print it */
      if(0 == XMLString::compareString(
	       instruction->getNodeName(),
	       XMLString::transcode("instruction"))
	||
	 0 == XMLString::compareString(
	       instruction->getNodeName(),
	       XSTR("pattern"))){
	target = instruction->getFirstChild();
	while(target){
	  if(0 == XMLString::compareString(target->getNodeName(),
					   XSTR("target"))
	    ||
	     0 == XMLString::compareString(target->getNodeName(),
					   XSTR("ptarget"))){
	    char *inst_id = XSTR(
		          ((DOMElement *)target)->getAttribute(
			  XSTR("id")));
	    char *inst_dest1 = XSTR(
			     ((DOMElement *)target)->getAttribute(
		             XSTR("dest1")));
	    char *inst_src1 = XSTR(
			     ((DOMElement *)target)->getAttribute(
		             XSTR("src1")));
	    char *inst_src2 = XSTR(
			     ((DOMElement *)target)->getAttribute(
		             XSTR("src2")));
	    if(0 == strcmp(inst_dest1, current_residence)){
	      prdest1.insert(inst_id);
	    }
	    
	    if (0 == strcmp(inst_src1, current_residence)){
	      prsrc1.insert(inst_id);
	    }
	    
	    if (0 == strcmp(inst_src2, current_residence)){
	      prsrc2.insert(inst_id);
	    }
	    XMLString::release(&inst_dest1);
	    XMLString::release(&inst_src1);
	    XMLString::release(&inst_src2);
	  }
	  target = target->getNextSibling();
	}
      }
      instruction = instruction->getNextSibling();
    }
    std::set< char * >::iterator i;
    _f << "set PRD[" << current_residence << "] :=" << std::endl;
    for(i=prdest1.begin(); i!=prdest1.end(); ++i){
      _f << "  " << *i << std::endl;
    }
    _f << ";" << std::endl;
    _f << "set PRS1[" << current_residence << "] :=" << std::endl;
    for(i=prsrc1.begin(); i!=prsrc1.end(); ++i){
      _f << "  " << *i << std::endl;
    }
    _f << ";" << std::endl;
    _f << "set PRS2[" << current_residence << "] :=" << std::endl;
    for(i=prsrc2.begin(); i!=prsrc2.end(); ++i){
      _f << "  " << *i << std::endl;
    }
    _f << ";" << std::endl;
    /* all inst_id are used, free them */
    for(i=prdest1.begin(); i!=prdest1.end(); ++i){
      char * inst_id = *i;
      XMLString::release(&inst_id);
    }
  }

  /* Transfers, every row is RS RS F 1 if the transfer uses F */
  _f << "param UX := " << std::endl;
  char *xsrc1, *xdest1, *resattr;
  DOMNode *transfer;
  DOMNode *cycle_matrix;
  DOMNode *target;
  transfer = get_sub_tree("transfer");
  /* If there are transfers */
  if(NULL != transfer){
    target = transfer->getFirstChild();
    while(NULL != target){
      if(0 == XMLString::compareString(
	       target->getNodeName(),
	       XMLString::transcode("target"))){
	xsrc1 = XSTR(((DOMElement *)target)->getAttribute(XSTR("src1")));
	xdest1 = XSTR(((DOMElement *)target)->getAttribute(XSTR("dest1")));
	cycle_matrix = get_sub_tree("cycle_matrix", target->getFirstChild());
	/* Loop over resources */
	std::map< char*, int, ltstr >::iterator resourceit;
	for(resourceit = stage_map.begin() ; resourceit!= stage_map.end(); 
	    ++resourceit){
	  if(((DOMElement *)cycle_matrix)->
	     hasAttribute(XSTR((*resourceit).first))){
	    /* This resource is used in the transfer */
	    _f << "  " << xsrc1 << " " << xdest1 
	       << " " << (*resourceit).first << " 1" << std::endl; 
	  }
	}
      }
      target = target->getNextSibling();
    }
  }
  _f << ";" << std::endl;

  /* Number of registers in each residence */
  _f << "param R := " << std::endl;
  DOMNode *residences = get_sub_tree("residences");
  DOMNode *residence = get_sub_tree("residence", residences->getFirstChild());
  while(residence){
    /* Filter out comments ? */
    if(0 == XMLString::compareString(
				     residence->getNodeName(),
				     XMLString::transcode("residence"))){
      for(residenceit = residence_map.begin(); residenceit != residence_map.end();
	  ++residenceit){
	if(0 == XMLString::
	   compareString(((DOMElement *)residence)->getAttribute(XSTR("id")),
			 XSTR((*residenceit).first))){
	  /* We have the right residence tree, now count */
	  int rescount = 0;
	  DOMNode *reg_pointer = residence->getFirstChild();
	  while(reg_pointer){
	    if(0 == XMLString::compareString(reg_pointer->getNodeName(),
					     XMLString::transcode("reg"))){
	      rescount++;
	    }
	    reg_pointer = reg_pointer->getNextSibling();
	  }
	  _f << "  " << (*residenceit).first << " " << rescount << std::endl;
	}
      }
    }
    residence = residence->getNextSibling();
  }
  _f << ";" << std::endl;

  /* Latencies of transfers */
  _f << "param LX :=" << std::endl;
  transfer = get_sub_tree("transfer");
  if(transfer){
  target = get_sub_tree("target", transfer->getFirstChild());
  while(target){
    if(0 == XMLString::compareString(target->getNodeName(), XSTR("target"))){
      /* Find latency */
      int latency_val = 0;
      DOMNode *ops = target->getFirstChild();
      int ell = 0;
      while(ops != NULL) {
	if(ops->getNodeType() == DOMNode::ELEMENT_NODE) {
	  if(!XMLString::compareString(ops->getNodeName(), 
				       XMLString::transcode("cycle_matrix"))) {
	    //Extract latency (ILP specific)
	    DOMNode *latency = ops->getFirstChild();
	    while(latency != NULL) {
	      if(!XMLString::compareString(latency->getNodeName(), 
					   XMLString::transcode("latency"))) {
		ell = max(ell,XMLString::parseInt(((DOMElement *)latency)
				->getAttribute(XMLString::transcode("l"))));
	      }
	      latency = latency->getNextSibling();
	    }
	    
	    DOMNamedNodeMap *anm = ops->getAttributes();
	    for(int i = 0; i < anm->getLength(); i++) {
	      DOMNode *att = anm->item(i);
	      if(att->getNodeType() != DOMNode::ATTRIBUTE_NODE) {
		continue;
	      }
	      char *stage_name = XMLString::transcode(att->getNodeName());
	      int length = XMLString::parseInt(((DOMAttr *)att)->getValue()) + ell;
	      
	      latency_val = length;
	    }
	  }
	}
	ops = ops->getNextSibling();
      }
     /* For now a transfer is identified by src1 and dest1, this will
	 probably change in the future */
      _f << "  " << XSTR(((DOMElement *)target)->getAttribute(XSTR("src1")))
	 <<  " " << XSTR(((DOMElement *)target)->getAttribute(XSTR("dest1")))
	 <<  " " << latency_val << std::endl;
    }
    target = target->getNextSibling();
  }
  }
  _f << std::endl << ";" << std::endl;

  _f << std::endl << "#end of archi::ilp" << std::endl;

  XMLString::release(&xinstruction);
  XMLString::release(&xtarget);
  XMLString::release(&xid);
  XMLString::release(&xcycle_matrix);
  XMLString::release(&xlatency);
  XMLString::release(&xl);
}

//Private methods
void Architecture::initialize(void)
{
  pattern_map.clear();
  residence_map.clear();
  op_map.clear();
  transfer_map.clear();
  stage_map.clear();
  remove_ops.clear();

  //Travers and build patterns
  build_stage_map();
  build_pattern_map();
  build_register_map();
  build_residence_map();
  //  build_fu_map();  //depricate

  //CONDITION --- CONDFILE is generated even if it is not used because
  //of the tight coupling with the following functions.
  std::ofstream file(CONDFILE);
  start_cond_file(file);
  build_op_map(file);
  build_transfer_map();
  end_cond_file(file);
  file.close();

  /* Only do this if there are conditions in the architecture
     specification. */
  if(cond_index > 0){
    /* Issue warning. (It might actually work, I'm not sure.) */
    std::cerr << "WARNING: unsupported feature: conditional instruction." << std::endl;

    //  system("gcc -I.../lcc-4.2/src -g -shared -o /tmp/cond.so /tmp/cond.cc")
    if(system(COMMAND)) {
      std::cerr << COMMAND << std::endl
		<< "Compilation of conditions failed!\n";
      exit(1);
    }

    handle = dlopen(CONDSO, RTLD_LAZY);
    if(!handle) {
      std::cerr << dlerror();
      exit(1);
    }
    //dlerror();
 
    conditions = (condition *)dlsym(handle, "conditions");
    char *error;
    if((error = dlerror()) != NULL) {
      std::cerr << error << std::endl;
      exit(1);
    }
  } /* ends if cond_index > 0 */

  build_remove_ops(); //<remove><op id = "1234"/><op id = "2345"/></remove>
}

//PATTERN
void Architecture::build_pattern_kid(DOMNode *_dom_node, int _kid_nr, 
				     char *_kid_id, pvertex_id & _v, pattern_t & _pattern)
{

  vertex_info_map::type vim = vertex_info_property(_pattern);
  vertex_kind_map::type vkm = vertex_kind_property(_pattern);
  pvertex_id u;
  bool b_found = false;
  
  if(strlen(_kid_id) > 0) {
    //lookup for an existing named kid (DAG patterns)
    pvertex_iter vi, vi_end;
    for(tie(vi, vi_end) = boost::vertices(_pattern);
	vi != vi_end; ++vi) {
      if(!strcmp(vim[*vi].id, _kid_id)) {
	//matches
	u = *vi;
	b_found = true;
	break;
      }
    }
  }

  if(strlen(_kid_id) == 0 || !b_found) {
    u = add_vertex(_pattern);
    vim[u].indeg = 0;
    vim[u].ops.clear();
    vim[u].id = _kid_id;
    vkm[u] = ARGNODE;
  }

  std::pair<pedge_id, bool> ep = add_edge(u, _v, _pattern);
  if(vkm[u] == OPNODE) {
    //in case of previous founded node
    vim[_v].indeg++;
  }
  pedge_type_map::type etm = edge_type_property(_pattern);
  assert(_kid_nr == 0 || _kid_nr == 1);
  etm[ep.first] = _kid_nr ? KID1 : KID0;


  while(_dom_node) {
    if(_dom_node->getNodeType() == DOMNode::ELEMENT_NODE) {

      if(!XMLString::compareString(_dom_node->getNodeName(), 
				   XMLString::transcode("node"))) {
	//Kid has explicit node
	build_pattern_node(_dom_node->getFirstChild(), u, _pattern);
	vim[_v].indeg++;
	vkm[u] = OPNODE;
      }

    }
    _dom_node = _dom_node->getNextSibling();
  }
}

void Architecture::build_pattern_node(DOMNode *_dom_node, pvertex_id & _v, pattern_t & _pattern)
{
  bool b_or = false;  //or section
  vertex_info_map::type vim = vertex_info_property(_pattern);
  int arity = 0;

  while(_dom_node || b_or) {
    if(_dom_node->getNodeType() == DOMNode::ELEMENT_NODE) {
      
      if(!XMLString::compareString(_dom_node->getNodeName(), 
				   XMLString::transcode("or"))) {
	_dom_node = _dom_node->getFirstChild();
	b_or = true;
      }
	 
      if(!XMLString::compareString(_dom_node->getNodeName(), 
				   XMLString::transcode("pinst"))) {

	//FIXME: to be removed
	//	const XMLCh * op_ch = ((DOMElement *)_dom_node)->getAttribute(XMLString::transcode("op"));
	//	int op_number = XMLString::parseInt(op_ch);
	//
	int op_number = get_op_number( XMLString::transcode(((DOMElement *)_dom_node)->getAttribute(XMLString::transcode("id"))));
	//	assert(op_number == opn);

	vim[_v].ops.push_back(op_number);
	if(arity == 0) {
	  arity = XMLString::parseInt(((DOMElement *)_dom_node)->getAttribute(XMLString::transcode("arity")));
	} else {
	  assert(arity == XMLString::parseInt(((DOMElement *)_dom_node)->getAttribute(XMLString::transcode("arity"))));
	}
      }
      
      if(!XMLString::compareString(_dom_node->getNodeName(), 
				   XMLString::transcode("kid"))) {
	int kid_nr = XMLString::parseInt(((DOMElement *)_dom_node)->getAttribute(XMLString::transcode("nr")));
	char *kid_id = XMLString::transcode(((DOMElement *)_dom_node)->getAttribute(XMLString::transcode("id")));
	build_pattern_kid(_dom_node->getFirstChild(), kid_nr, kid_id, _v, _pattern);
      }
    }
    if((_dom_node->getNextSibling() == NULL) && b_or) {
      _dom_node = _dom_node->getParentNode();
      b_or = false;
    }
    _dom_node = _dom_node->getNextSibling();
  }
}

void Architecture::build_tree_pattern(DOMNode *_dom_node, pattern_t & _pattern)
{
  typedef std::pair< enum EDGE_T, std::pair< char *, char * > > dep_edge_t;
  std::vector< std::pair< enum EDGE_T, std::pair< char *, char * > > > deps;
  vertex_kind_map::type vkm = vertex_kind_property(_pattern);
  while(_dom_node) {
    if(_dom_node->getNodeType() == DOMNode::ELEMENT_NODE) {
      if(!XMLString::compareString(_dom_node->getNodeName(), 
				   XMLString::transcode("node"))) {
	pvertex_id v = add_vertex(_pattern);
	vkm[v] = OPNODE;
	vertex_info_map::type vim = vertex_info_property(_pattern);
	vim[v].indeg = 0;
	vim[v].ops.clear();
	vim[v].id = XMLString::transcode(((DOMElement *)_dom_node)->getAttribute(XMLString::transcode("id")));
	build_pattern_node(_dom_node->getFirstChild(), v, _pattern);
      }

      if(!XMLString::compareString(_dom_node->getNodeName(), 
				   XMLString::transcode("ddep")) ||
	 !XMLString::compareString(_dom_node->getNodeName(),
				   XMLString::transcode("link"))) {
	deps.push_back(dep_edge_t());
	dep_edge_t & de = deps.back();
	de.first = (!XMLString::compareString(_dom_node->getNodeName(),
					      XMLString::transcode("ddep"))) ? DDEP : LINK;
	de.second.first = XMLString::transcode(((DOMElement *)_dom_node)->getAttribute(XMLString::transcode("src")));
	de.second.second = XMLString::transcode(((DOMElement *)_dom_node)->getAttribute(XMLString::transcode("dest")));;
      }
    }
    _dom_node = _dom_node->getNextSibling();
  }

  if(deps.size() > 0) {
    //Additional dependencies
    pedge_type_map::type etm = edge_type_property(_pattern);
    vertex_info_map::type vim = vertex_info_property(_pattern);

    for(int i = 0; i < deps.size(); i++) {
      bool b_src = false, b_dest = false;
      pvertex_id src;
      pvertex_id dest;
      pvertex_iter vi, vi_end;

      for(tie(vi, vi_end) = boost::vertices(_pattern); vi != vi_end; ++vi) {
	if(!strcmp(vim[*vi].id, deps[i].second.first)) {
	  src = *vi;
	  b_src = true;
	}
	if(!strcmp(vim[*vi].id, deps[i].second.second)) {
	  dest = *vi;
	  b_dest = true;
	}
      }
      if(b_src && b_dest) {
	std::pair< pedge_id, bool > dep_edge = add_edge(src, dest, _pattern);
	etm[dep_edge.first] = deps[i].first;
	if(vkm[src] == OPNODE) {
	  vim[dest].indeg++;
	}
      } else {
	std::cerr << "Dependece edge has unspecified src: " 
		  << deps[i].second.first << " " << b_src << " /and destination " 
		  << deps[i].second.second << " " << b_dest << std::endl;
	exit(-1);
      }
    }
  }
}

void Architecture::build_pattern_map()
{
  DOMNode *patterns = get_sub_tree("patterns");

  //If the specification contains no patterns
  if(!patterns) {
    return;
  }
  DOMNode *dom_node = patterns->getFirstChild();

  char *pattern_id;
  char file_name[55];

  while(dom_node) {
    if(dom_node->getNodeType() == DOMNode::ELEMENT_NODE) {
      if(!XMLString::compareString(dom_node->getNodeName(), XMLString::transcode("pattern"))) {
	char *pattern_id = XMLString::transcode(((DOMElement *)dom_node)->getAttribute(XMLString::transcode("id")));
	pattern_t *pattern = new pattern_t;
	pattern_map[pattern_id] = pattern;
	build_tree_pattern(dom_node->getFirstChild(), *pattern);

	if(Verbose::verbosityLevel > 2) {
	  //debug
	  if(strlen(pattern_id) < 50) {
	    sprintf(file_name, "pat_%s.%s", pattern_id, "dot");
	  } else {
	    std::cerr << "pattern name: " << pattern_id 
		      << " too long for dumping (FIXME)!\n";
	  }

	  std::ofstream dot_file(file_name);
	  pnode_writer< pattern_t & > nw(*pattern);
	  pedge_writer< pattern_t & > ew(*pattern);
	  write_graphviz(dot_file, *pattern, nw, ew);
	  dot_file.close();
	  // end debug
	}
      }
    }
    dom_node = dom_node->getNextSibling();
  }
}

void Architecture::build_register_map(void)
{
  char *rid;
  int size;
  int index = 0;
  XMLCh * regx = XMLString::transcode("reg");
  XMLCh * idx = XMLString::transcode("id");
  XMLCh * sizex = XMLString::transcode("size");

  DOMNode *regs = get_sub_tree("registers");

  DOMNode *regCursor = regs->getFirstChild();
  while(regCursor != NULL) {
    if(regCursor->getNodeType() == DOMNode::ELEMENT_NODE &&
       !XMLString::compareString(regCursor->getNodeName(), regx)) {
      rid = XMLString::transcode(((DOMElement *)regCursor)->getAttribute(idx));
      size = -1;
      const XMLCh *pSize = ((DOMElement *)regCursor)->getAttribute(sizex);
      
      XMLCh * empty = XMLString::transcode("");
      if(!XMLString::equals(pSize, empty)) {
	size = XMLString::parseInt(pSize);
      }
      if(empty) XMLString::release(&empty);
      register_map[rid].size = size;
      register_map[rid].index = index;
      register_index[index] = rid;
      index++;
    }
    regCursor = regCursor->getNextSibling();
  }
  XMLString::release(&regx);
  XMLString::release(&idx);
  XMLString::release(&sizex);
}

void Architecture::build_residence_map(void)
{
  DOMNode *residences = get_sub_tree("residences");
  assert(residences);
  DOMNode *residence = residences->getFirstChild();
  int index = 0;
  bool error = false;

  while(residence != NULL) {
    if(residence->getNodeType() == DOMNode::ELEMENT_NODE &&
       !XMLString::compareString(residence->getNodeName(), XMLString::transcode("residence"))) {
      char *res_id = XMLString::transcode(((DOMElement *)residence)->getAttribute(XMLString::transcode("id")));
      rc_resources.push_back(std::vector< int >());
      rc_resources[index].push_back(-1);  //For unspecifed #reg
      residence_index[res_id] = index++;
      assert(rc_resources.size() == index);
      std::vector< int > & tmp_resources = rc_resources.back();
      assert(tmp_resources.size() == 1);
      DOMNode *reg = residence->getFirstChild();
      bool b_first = true;
      all_residences.insert(res_id);

      while(reg != NULL) {
	if(reg->getNodeType() == DOMNode::ELEMENT_NODE &&
	   !XMLString::compareString(reg->getNodeName(), XMLString::transcode("reg"))) {
	  char *reg_id = XMLString::transcode(((DOMElement *)reg)->getAttribute(XMLString::transcode("id")));
	  if(register_map.find(reg_id) == register_map.end()) {
	    std::cerr << "Resource: '" << reg_id 
		      << "' is not declared, but used to build residence class: '" 
		      << res_id << "'" << std::endl;
	    error = true;
	    
	  }
	  residence_map[res_id].push_back(register_map[reg_id].index);
	  if(b_first) {
	    tmp_resources[0] = (register_map[reg_id].index);
	    b_first = false;
	  } else {
	    tmp_resources.push_back(register_map[reg_id].index);
	  }
	}
	reg = reg->getNextSibling();
      }
    }
    residence = residence->getNextSibling();
  }
  if(error) {
    std::cerr << "Failed to build residences.\n";
    exit(-1);
  }
  //DEBUG
  //for(int j = 0; j < rc_resources.size(); j++) {
  //	std::cout << j << ": ";
  //	std::vector<int> tmp = rc_resources[j];
  //	for(int i = 0; i < tmp.size(); i++) {
  //		std::cout << tmp[i] << " ";
  //	}
  //	std::cout << std::endl;
  // }
  //end debug
}


void Architecture::build_reserv_tab(DOMNode *_reserv, reserv_tab_t  & _reserv_tab)
{
  assert(_reserv_tab.size() == stage_map.size());
  while(_reserv != NULL) {
    if(_reserv->getNodeType() == DOMNode::ELEMENT_NODE) {

      if(!XMLString::compareString(_reserv->getNodeName(), XMLString::transcode("latency"))) {
	int ell = XMLString::parseInt(((DOMElement *)_reserv)->getAttribute(XMLString::transcode("l")));
	int max_height = 0;
	for(int p = 0; p < _reserv_tab.size(); p++) {
	  max_height = max(max_height, _reserv_tab[p].size());
	}
	//Add latency delay
	for(int p = 0; p < _reserv_tab.size(); p++) {
	  while(_reserv_tab[p].size() < max_height + ell) {
	    _reserv_tab[p].push_back(false);
	  }
	}
      } else if(!XMLString::compareString(_reserv->getNodeName(), XMLString::transcode("cycle_matrix"))) {
	
	DOMNamedNodeMap *anm = _reserv->getAttributes();

	/*
	  if(_reserv_tab->size() == 0) {
	  //new
	  for(int i = 0; i < anm->getLength(); i++) {
	  DOMNode *att = anm->item(i);
	  if(att->getNodeType() != DOMNode::ATTRIBUTE_NODE) {
	  continue;
	  }
	  char *stage_name = XMLString::transcode(att->getNodeName());
	  int length = XMLString::parseInt(((DOMAttr *)att)->getValue());
	  std::map< char *, int, ltstr >::iterator smi = stage_map.find(stage_name);
	  if(smi == stage_map.end()) {
	  //new entry
	  stage_map[stage_name] = stage_map.size();
	  }
	  int stage_number = stage_map[stage_name];
	  while(_reserv_tab->size() <= stage_number) {
	  _reserv_tab->push_back(bit_vector_t(0));
	  }
	  for(int j = 0; j < length; j++) {
	  (*_reserv_tab)[stage_number].push_back(true);
	  }
	  }
	  } else {
	*/
	//Compute current max hight
	int max_height = 0;
	for(int p = 0; p < _reserv_tab.size(); p++) {
	  max_height = max(max_height, _reserv_tab[p].size());
	}
	for(int i = 0; i < anm->getLength(); i++) {
	  DOMNode *att = anm->item(i);
	  if(att->getNodeType() != DOMNode::ATTRIBUTE_NODE) {
	    continue;
	  }
	  char *stage_name = XMLString::transcode(att->getNodeName());
	  int length = XMLString::parseInt(((DOMAttr *)att)->getValue());
	  std::map< char *, int, ltstr >::iterator smi = stage_map.find(stage_name);
	  assert(smi != stage_map.end());
	  int column = smi->second;
	  int column_height = _reserv_tab[column].size();
	  //	std::cout << "\nColumn height: " << column_height << std::endl;
	  //	std::cout << "\nReserv tab size: " << (*_reserv_tab).size();
	  for(int j = column_height; j < max_height; j++) {
	    //Padd with empty entries
	    _reserv_tab[column].push_back(false);
	  }
	
	  //append
	  for(int k = 0; k < length; k++) {
	    _reserv_tab[column].push_back(true);
	  }
	}
	if(_reserv->hasChildNodes()) {
	  build_reserv_tab(_reserv->getFirstChild(), _reserv_tab);
	}
      }
    }
    _reserv = _reserv->getNextSibling();
  }
}

int Architecture::build_target_instruction(DOMNode * _target, 
					   std::vector< TargetInstruction * > & _target_instrs,
					   std::ofstream & _f)
{
  int nb_op = 0;
  while(_target != NULL) {
    if(_target->getNodeType() == DOMNode::ELEMENT_NODE &&
       !XMLString::compareString(_target->getNodeName(), XMLString::transcode("target"))) {
      char *mnemonic = XMLString::transcode(((DOMElement *)_target)->getAttribute(XMLString::transcode("id")));

      //Functional unit
      std::vector< int > fus(0);
      float bcost = 0.0;
      get_fu_data(_target, &bcost, fus);

      char *dest1 = XMLString::transcode(((DOMElement *)_target)->getAttribute(XMLString::transcode("dest1")));
      int i_dest1 = STMT_RC; //no dest1
      if(strlen(dest1) > 0) {
	if(residence_index.find(dest1) == residence_index.end()) {
	  std::cerr << "Residence: " << dest1 << " does not exist target: " << mnemonic << std::endl;
	  exit(-1);
	}
	i_dest1 = residence_index[dest1];
      }
      XMLString::release(&dest1);
      nb_op = 1;
      int i_src1 = -1, i_src2 = -1;
      char *src1 = XMLString::transcode(((DOMElement *)_target)->getAttribute(XMLString::transcode("src1")));
      if(strlen(src1) > 0) {
	if(residence_index.find(src1) == residence_index.end()) {
	  std::cerr << "Residence: " << src1 << " does not exist target: " << mnemonic << std::endl;
	  exit(-1);
	}
	i_src1 = residence_index[src1];
	nb_op = 2;
	char *src2 = XMLString::transcode(((DOMElement *)_target)->getAttribute(XMLString::transcode("src2")));
	if(strlen(src2) > 0) {
	  if(residence_index.find(src2) == residence_index.end()) {
	    std::cerr << "Residence: " << src2 << " does not exist target: " << mnemonic << std::endl;
	    exit(-1);
	  }
	  i_src2 = residence_index[src2];
	  nb_op = 3;
	}
	XMLString::release(&src1);
	XMLString::release(&src2);
      }
      //reservation table
      reserv_tab_t *reserv_tab = new reserv_tab_t(stage_map.size());
      build_reserv_tab(_target->getFirstChild(), *reserv_tab);
      build_fus(_target->getFirstChild(), fus);
      char *format = "";
      build_format(_target->getFirstChild(), format);
      char *constraint = 0;
      build_constraint(_target->getFirstChild(), constraint);
      TargetInstruction *target_instr = new TargetInstruction(mnemonic, reserv_tab, i_dest1, 
							      i_src1, i_src2, 
							      format, bcost, fus, constraint);
      _target_instrs.push_back(target_instr);

      //CONDITION
      build_target_condition(_target->getFirstChild(), _f, target_instr);
      //
      XMLString::release(&mnemonic);
    }
    _target = _target->getNextSibling();
  }
  return nb_op;
}

void Architecture::build_fus(DOMNode *_functional_units, std::vector< int > & _fus)
{
  while(_functional_units) {
    if(_functional_units->getNodeType() == DOMNode::ELEMENT_NODE &&
       !XMLString::compareString(_functional_units->getNodeName(), XMLString::transcode("use_fu"))) {

      const XMLCh *use_fu = ((DOMElement *)_functional_units)->getAttribute(XMLString::transcode("id"));
      if(!XMLString::equals(use_fu, XMLString::transcode(""))) {
	char *fu = XMLString::transcode(use_fu);
	if(funits_map.find(fu) == funits_map.end()) {
	  std::cerr << "FU: " << fu << " undefined!\n";
	  assert(false);
	}
	_fus.push_back(funits_map[fu]);
      }
    }
    _functional_units = _functional_units->getNextSibling();
  }
}

void Architecture::build_op_map_instruction(DOMNode *_instruction, std::ofstream & _f)
{
  XMLCh * idx = XMLString::transcode("id");
  //build pattern (single node)
  int op_number = get_op_number(XMLString::transcode(((DOMElement *)_instruction)->getAttribute(idx)));
  char *instr_id = XMLString::transcode(((DOMElement *)_instruction)->getAttribute(idx));
  XMLString::release(&idx);
  pattern_t *instr_pattern = new pattern_t;
  pvertex_id v = add_vertex(*instr_pattern);
  vertex_kind_map::type vkm = vertex_kind_property(*instr_pattern);
  vkm[v] = OPNODE;
  vertex_info_map::type vim = vertex_info_property(*instr_pattern);
  vim[v].indeg = 0;
  vim[v].ops.clear();
  vim[v].id = instr_id;
  vim[v].ops.push_back(op_number);

  std::vector< TargetInstruction * > target_instrs;
  int nb_op = build_target_instruction(_instruction->getFirstChild(), target_instrs, _f);

  if(nb_op > 1) {
    vertex_info_map::type vim = vertex_info_property(*instr_pattern);
    pedge_type_map::type etm = edge_type_property(*instr_pattern);
    vertex_kind_map::type vkm = vertex_kind_property(*instr_pattern);
    switch(nb_op) {

    case 3: {
      pvertex_id kid1 = add_vertex(*instr_pattern);
      std::pair< pedge_id, bool > kid1_e = add_edge(kid1, v, *instr_pattern);
      vim[kid1].indeg = 0;
      vim[kid1].ops.clear();
      vim[kid1].id = "src2";
      vim[kid1].arg = 1;  //src2 is args[1]
      etm[kid1_e.first] = KID1;
      vkm[kid1] = ARGNODE;
    }

    case 2: {
      pvertex_id kid0 = add_vertex(*instr_pattern);
      std::pair< pedge_id, bool > kid0_e = add_edge(kid0, v, *instr_pattern);
      vim[kid0].indeg = 0;
      vim[kid0].ops.clear();
      vim[kid0].id = "src1";
      vim[kid0].arg = 0;  //src1 is args[0]
      etm[kid0_e.first] = KID0;
      vkm[kid0] = ARGNODE;
    }
    }
  }

  if(Verbose::verbosityLevel > 2) {
    //debug
    static int variant;
    char file_name[256];
    sprintf(file_name, "%d_%d.%s", op_number, variant++, "dot");
      
    std::ofstream dot_file(file_name);
    pnode_writer< pattern_t & > nw(*instr_pattern);
    pedge_writer< pattern_t & > ew(*instr_pattern);
    write_graphviz(dot_file, *instr_pattern, nw, ew);
    dot_file.close();
    // end debug
  }

  for(int i = 0; i < target_instrs.size(); i++) {
    pattern_instr_t * pi = new pattern_instr_t;
    pi->first = instr_pattern;
    pi->second = target_instrs[i];
    op_map[op_number].push_back(pi);
  }
}

void Architecture::build_id_rclass_map(DOMNode *ops, 
				       std::map< char *, int, ltstr > & _id_rclass_map,
				       reserv_tab_t *_reserv_tab)
{
  while(ops != NULL) {
    if(ops->getNodeType() == DOMNode::ELEMENT_NODE) {

      if(!XMLString::compareString(ops->getNodeName(), XMLString::transcode("op"))) {
	char *op_id = XMLString::transcode(((DOMElement *)ops)->getAttribute(XMLString::transcode("id")));
	char *rc_id_name;
	DOMNode *rc_id = ops->getFirstChild();

	while(rc_id != NULL) {
	  if(rc_id->getNodeType() == DOMNode::ELEMENT_NODE &&
	     !XMLString::compareString(rc_id->getNodeName(), XMLString::transcode("id"))) {
	    rc_id_name = XMLString::transcode(rc_id->getFirstChild()->getNodeValue());
	    break; //no or statement
	  }
	  rc_id = rc_id->getNextSibling();
	}
	assert(rc_id_name);
	if(residence_index.find(rc_id_name) == residence_index.end()) {
	  std::cerr << "Residence class " << rc_id_name << " not found for pattern instruction!\n";
	  exit(-1);
	}
	_id_rclass_map[op_id] = residence_index[rc_id_name];
      }

      if(!XMLString::compareString(ops->getNodeName(), XMLString::transcode("cycle_matrix"))) {
	build_reserv_tab(ops, *_reserv_tab);
      }

    }
    ops = ops->getNextSibling();
  }
}

void Architecture::check_match(pattern_t *_pattern,
			       std::map< char *, int, ltstr > _id_rclass_map,
			       std::vector< int > & op_numbers)
{
  pvertex_iter pvi, pvi_end;
  vertex_info_map::type vim = vertex_info_property(*_pattern);
  vertex_kind_map::type vkm = vertex_kind_property(*_pattern);

  for(tie(pvi, pvi_end) = boost::vertices(*_pattern);
      pvi != pvi_end; ++pvi) {
    if(vkm[*pvi] == ARGNODE) {
      char *argnode_name = vim[*pvi].id;
      if(_id_rclass_map.find(argnode_name) == _id_rclass_map.end()) {
	std::cerr << "Mapping missing for ARGNODE: " << argnode_name << " for pattern!\n";
	exit(-1);
      }
    } else {
      assert(vkm[*pvi] == OPNODE);
      //opnode has no parameter
      if(vim[*pvi].indeg == 0) {
	std::list< int >::iterator opi = vim[*pvi].ops.begin(),
	  opi_end = vim[*pvi].ops.end();
	for(; opi != opi_end; ++opi) {
	  op_numbers.push_back(*opi);
	}
      }
    }
  }
}

void Architecture::build_op_map_pattern(DOMNode *_pattern, std::ofstream & _f)
{
  //patterns are built ahead pattern_map id->pattern (graph)
  char *pattern_id = XMLString::transcode(((DOMElement *)_pattern)->getAttribute(XMLString::transcode("id")));
  if(pattern_map.find(pattern_id) == pattern_map.end()) {
    std::cerr << "Pattern: " << pattern_id << " not defined!" << std::endl;
    exit(-1);
  }
  pattern_t *pattern = pattern_map[pattern_id];
  //Build ptarget
  DOMNode *ptarget = _pattern->getFirstChild();
  while(ptarget != NULL) {
    if(ptarget->getNodeType() == DOMNode::ELEMENT_NODE && 
       !XMLString::compareString(ptarget->getNodeName(), XMLString::transcode("ptarget"))) {
      char *mnemonic = XMLString::transcode(((DOMElement *)ptarget)->getAttribute(XMLString::transcode("id")));
      char *dest1_rc = XMLString::transcode(((DOMElement *)ptarget)->getAttribute(XMLString::transcode("dest1")));

      std::map< char *, int, ltstr > id_rclass_map;

      if(strlen(dest1_rc) == 0) {
	id_rclass_map["dest1"] = STMT_RC;
      } else {
	if(residence_index.find(dest1_rc) == residence_index.end()) {
	  std::cerr << "Residence class: " << dest1_rc << " undefined in pattern instruction: " << mnemonic << std::endl;
	  exit(-1);
	}
	id_rclass_map["dest1"] = residence_index[dest1_rc];
      }

      reserv_tab_t *reserv_tab = new reserv_tab_t(stage_map.size());
      build_id_rclass_map(ptarget->getFirstChild(), id_rclass_map, reserv_tab);

      char *format = "";
      build_format(ptarget->getFirstChild(), format);

      char *constraint = 0;
      build_constraint(ptarget->getFirstChild(), constraint);

      //Functional unit
      std::vector< int > fus(0);
      float bcost = 0.0;
      get_fu_data(ptarget, &bcost, fus);

      //check with the pattern and extracts ops (OPNODE && inded == 0) std::vector< int >
      std::vector< int > op_numbers;
      op_numbers.clear();

      pattern_t *_pattern_arg = pattern;
      std::map< char *, int, ltstr > _id_rclass_map_arg =  id_rclass_map;
      std::vector< int > & op_numbers_arg = op_numbers;

      check_match(_pattern_arg, _id_rclass_map_arg, op_numbers_arg);
      //      check_match(pattern, id_rclass_map, op_numbers);

      for(int i = 0; i < op_numbers.size(); i++) {
	//build instruction
	TargetInstruction *t_instr = new TargetInstruction(mnemonic, reserv_tab, id_rclass_map, format, bcost, fus, constraint);
	//debug
	//	std::cout << "Build2 TI: " << *t_instr << std::endl;
	//
	pattern_instr_t *pi = new pattern_instr_t;
	pi->first = pattern;
	pi->second = t_instr;
	op_map[ op_numbers[i] ].push_back(pi);

	//CONDITION
	build_target_condition(ptarget->getFirstChild(), _f, t_instr, pi->first);
      }
    }
    ptarget = ptarget->getNextSibling();
  }
}

void Architecture::build_format(DOMNode *_target, char * &  _format)
{
  _format = ""; //Shall we apply here the default format? We need the operands in that  case.
  while(_target != NULL) {
    if(_target->getNodeType() == DOMNode::ELEMENT_NODE &&
       !XMLString::compareString(_target->getNodeName(), XMLString::transcode("format"))) {
      _format = XMLString::transcode(_target->getFirstChild()->getNodeValue());
      break; //single format per instruction target
    }
    _target = _target->getNextSibling();
  }
  if(strlen(_format) == 0) { 
    _format = "";
  }
}

void Architecture::build_constraint(DOMNode *_target, char * &  _constraint)
{
  _constraint = 0; //No constraint
  while(_target != NULL) {
    if(_target->getNodeType() == DOMNode::ELEMENT_NODE &&
       !XMLString::compareString(_target->getNodeName(), XMLString::transcode("constraint"))) {
      _constraint = XMLString::transcode(((DOMElement *)_target)->getAttribute(XMLString::transcode("dest1")));
      break; //single constraint per instruction target
    }
    _target = _target->getNextSibling();
  }
}

void Architecture::build_op_map(std::ofstream & _f)
{
  DOMNode *instructionset = get_sub_tree("instructionset");
  assert(instructionset);
  DOMNode *instruction = instructionset->getFirstChild();

  while(instruction != NULL) {
    if(instruction->getNodeType() == DOMNode::ELEMENT_NODE) {
      if(!XMLString::compareString(instruction->getNodeName(), XMLString::transcode("instruction"))) {
	build_op_map_instruction(instruction, _f);
      }
      if(!XMLString::compareString(instruction->getNodeName(), XMLString::transcode("pattern"))) {
	build_op_map_pattern(instruction, _f);
      }
    }
    instruction = instruction->getNextSibling();
  }
}

void Architecture::build_transfer_map(void)
{
  DOMNode *transfer = get_sub_tree("transfer");
  if(!transfer) {
    return;
  }
  DOMNode *target = transfer->getFirstChild();
  while(target != NULL) {
    if(target->getNodeType() == DOMNode::ELEMENT_NODE &&
       !XMLString::compareString(target->getNodeName(), XMLString::transcode("target"))) {

      char *transfer_id = XMLString::transcode(((DOMElement *)target)->getAttribute(XMLString::transcode("id")));
      char *dest_rc = XMLString::transcode(((DOMElement *)target)->getAttribute(XMLString::transcode("dest1")));
      char *src_rc = XMLString::transcode(((DOMElement *)target)->getAttribute(XMLString::transcode("src1")));
      assert((strlen(dest_rc) > 0) && (strlen(src_rc) > 0));

      //Functional unit
      std::vector< int > fus(0);
      float bcost = 0.0;
      get_fu_data(target, &bcost, fus);

      //Check if the residence are defined for source and destination
      if(residence_index.find(dest_rc) == residence_index.end()) {
	std::cerr << "Residence: " << dest_rc << " does not exist target: " << transfer_id << std::endl;
	exit(-1);
      }
      if(residence_index.find(src_rc) == residence_index.end()) {
	std::cerr << "Residence: " << src_rc << " does not exist target: " << transfer_id << std::endl;
	exit(-1);
      }
      int i_dest = residence_index[dest_rc];
      int i_src = residence_index[src_rc];

      //One reservation table and one format per transfer instruction
      DOMNode *rtf = target->getFirstChild(); //Reservation table and format
      bool b_reservation_table = false, b_format = false;
      char *format = "";
      reserv_tab_t *reserv_tab = new reserv_tab_t(stage_map.size());
      assert(reserv_tab->size() == stage_map.size());
      while(rtf != NULL) {
	if(rtf->getNodeType() == DOMNode::ELEMENT_NODE) {
	  if(!XMLString::compareString(rtf->getNodeName(), XMLString::transcode("cycle_matrix"))) {
	    assert(!b_reservation_table);
	    b_reservation_table = true;
	    build_reserv_tab(rtf, *reserv_tab);
	  } else if(!XMLString::compareString(rtf->getNodeName(), XMLString::transcode("format"))) {
	    assert(!b_format);
	    b_format = true;
	    build_format(rtf, format);
	  }
	}
	rtf = rtf->getNextSibling();
      }

      //Build transfer
      TargetInstruction *target_instr = new TargetInstruction(transfer_id, reserv_tab, i_dest, i_src, -1, format, bcost, fus);
      transfer_map[i_src].push_back(std::make_pair(i_dest, target_instr));
    }
    target = target->getNextSibling();
  }
}

std::pair< bool,
	   std::vector< std::pair< int, TargetInstruction * > > >
Architecture::get_transfer_destinations(int _source)
{
  if(transfer_map.find(_source) == transfer_map.end()) {
    return std::make_pair(false, std::vector< std::pair< int, TargetInstruction * > >());
  }
  return std::make_pair(true, transfer_map[_source]);
}

void Architecture::build_stage_map(void)
{
  DOMNode *p = get_sub_tree("resources");
  if(!p) {
    std::cerr << "<resources> section not found!\n";
    exit(-1);
  } else {
    //Build the table of resources
    p = p->getFirstChild();
    while(p) {
      XMLCh * tmp = XMLString::transcode("resource");
      if(!XMLString::compareString(p->getNodeName(), tmp )) {
	//Insert resource name into stage map
	char *resource_id = XMLString::transcode(((DOMElement *)p)->getAttribute(XMLString::transcode("id")));
	if(stage_map.find(resource_id) != stage_map.end()) {
	  std::cerr << "Resource: " << resource_id << " is declared twice!\n";
	  exit(-1);
	} else {
	  int index = stage_map.size();
	  stage_map.insert(std::make_pair(resource_id, index));
	}
      }
      if(tmp) XMLString::release(&tmp);
      p = p->getNextSibling();
    }
  }

  //Filter: error checking for reservation tables
  DOMNode *root = spec->getDocumentElement();
  DOMTreeWalker *walker = spec->createTreeWalker(root, 
						 DOMNodeFilter::SHOW_ELEMENT, NULL, true);
  for(DOMNode *node = walker->nextNode(); node != NULL; node = walker->nextNode()) {
    if(!XMLString::compareString(node->getNodeName(), XMLString::transcode("cycle_matrix"))) {
      DOMNamedNodeMap *anm = node->getAttributes();
      for(int i = 0; i < anm->getLength(); i++) {
	DOMNode *att = anm->item(i);
	if(att->getNodeType() != DOMNode::ATTRIBUTE_NODE) {
	  continue;
	}
	char *stage_name = XMLString::transcode(att->getNodeName());
	if(stage_map.find(stage_name) == stage_map.end()) {
	  std::cerr << "Resource: " 
		    << stage_name 
		    << " is used in cycle_matrix (reservation table)\n"
		    << "without being declared in the resources section!\n";
	  exit(-1);
	  //stage_map.insert(std::make_pair(stage_name, stage_map.size()));
	}
      }
    }
  }

  //debug
  //   std::cout << "\nDUMP STAGE MAP\n";
  //   for(std::map< char *, int, ltstr >::const_iterator iter = stage_map.begin();
  //       iter != stage_map.end(); ++iter){
  //     std::cout << iter->first << ": " << iter->second << std::endl;
  //   }
  //
}

void Architecture::build_fu_map(void)
{
  DOMNode *funits = get_sub_tree("funits");

  if(!funits) {
    std::cerr << "Warning: section funits missing in the hardware specifications!\n";
    return;
  }
  DOMNode *funit = funits->getFirstChild();

  while(funit) {
    if(funit->getNodeType() == DOMNode::ELEMENT_NODE &&
       !XMLString::compareString(funit->getNodeName(), XMLString::transcode("fu"))) {
      char *id = XMLString::transcode(((DOMElement *)funit)->getAttribute(XMLString::transcode("id")));

      int occupation = 0;
      const XMLCh *p_occupation = ((DOMElement *)funit)->getAttribute(XMLString::transcode("occupation"));
      if(!XMLString::equals(p_occupation, XMLString::transcode(""))) {
	occupation = XMLString::parseInt(p_occupation);
      }

      int latency = 0;
      const XMLCh *p_latency = ((DOMElement *)funit)->getAttribute(XMLString::transcode("latency"));
      if(!XMLString::equals(p_latency, XMLString::transcode(""))) {
	latency = XMLString::parseInt(p_latency);
      }

      float activation_cost = 0.0;
      const XMLCh *p_acost = ((DOMElement *)funit)->getAttribute(XMLString::transcode("acost"));
      if(!XMLString::equals(p_acost, XMLString::transcode(""))) {
	char *endptr;
	activation_cost = strtof(XMLString::transcode(p_acost), &endptr);
      }

      funits_map[id] = fu_id_map.size();
      fu_id_map.push_back(fu_data_s());
      fu_data_s & tmp_data = fu_id_map.back();
      tmp_data.id = id;
      tmp_data.occupation = occupation;
      tmp_data.latency = latency;
      tmp_data.activation_cost = activation_cost;
    }
    funit = funit->getNextSibling();
  }
}

int Architecture::get_funits(void)
{
  return fu_id_map.size();
}

float Architecture::get_activation_cost(int _fu)
{
  assert(_fu < fu_id_map.size());
  return fu_id_map[_fu].activation_cost;
}

void Architecture::get_fu_data(DOMNode *target, float *_b_cost, std::vector< int > & _fus)
{
  //Functional unit
  const XMLCh *use_fu = ((DOMElement *)target)->getAttribute(XMLString::transcode("use_fu"));
  if(!XMLString::equals(use_fu, XMLString::transcode(""))) {
    char *fu = XMLString::transcode(use_fu);

    if(funits_map.find(fu) == funits_map.end()) {
      std::cerr << "FU: " << fu << " undefined!\n";
      assert(false);
    }
    _fus.push_back(funits_map[fu]);
  }
  build_fus(target->getFirstChild(), _fus);

  //Base cost
  const XMLCh *p_bcost = ((DOMElement *)target)->getAttribute(XMLString::transcode("bcost"));
  if(!XMLString::equals(p_bcost, XMLString::transcode(""))) {
    char *endptr;
    *_b_cost = strtof(XMLString::transcode(p_bcost), &endptr);
  }
}

void Architecture::build_remove_ops(void)
{
  DOMNode *remove = get_sub_tree("remove");
  if(remove) {
    DOMNode *op = remove->getFirstChild();
    while(op != NULL) {
      if(op->getNodeType() == DOMNode::ELEMENT_NODE &&
	 !XMLString::compareString(op->getNodeName(), XMLString::transcode("op"))) {
	int i_op = get_op_number(XMLString::transcode(((DOMElement *)op)->getAttribute(XMLString::transcode("id"))));
	remove_ops.insert(i_op);
      }
      op = op->getNextSibling();
    }
  }
}

void Architecture::build_target_condition(DOMNode * _instruction, std::ofstream & _f, TargetInstruction *_target_instr, pattern_t *_pattern)
{
  static int index;
  while(_instruction != NULL) {
    if(_instruction->getNodeType() == DOMNode::ELEMENT_NODE) {
      char *str = XMLString::transcode(_instruction->getNodeName());
      //      std::cout << str << std::endl;
      if(!XMLString::compareString(_instruction->getNodeName(), XMLString::transcode("condition"))) {
	//The entry in the condition table
	_target_instr->set_cond_index(index);
	_f << "// " << _target_instr->get_mnemonic() << std::endl
	   << "static bool __cond_" << index++ << "(std::vector< lcc::Node > _args) {\n\treturn (";
	//	std::stack< int > logic_op;
	//	b_out = false;
	//	C_condition(_instruction->getFirstChild(), _f, logic_op, _pattern);
	std::stack< std::pair< bool, bool > > logic_op;
	//See cond(...) method for initial parameter of logical stack
	logic_op.push(std::make_pair(true, false));
	cond(_instruction->getFirstChild(), _f, logic_op, _pattern);
	_f << ");\n}\n";
      }
    }
    _instruction = _instruction->getNextSibling();
  }
  cond_index = index;
}

// Stack: 1st parameter of the pair represents the logical operation true - AND, false - OR
// The 2nd component is the memory if the operator should be print or not. Print if true, otherwise continue
// The stack is intialiazed with <true, false>: default AND
void Architecture::cond(DOMNode *_condition, std::ofstream & _f, 
			std::stack< std::pair< bool, bool > > & _logic_op,
			pattern_t *_pattern)
{

  bool AND = true;
  bool OR = false;

  while(_condition != NULL) {
    if(_condition->getNodeType() == DOMNode::ELEMENT_NODE &&
       !XMLString::compareString(_condition->getNodeName(), XMLString::transcode("and"))) {
      _logic_op.push(std::make_pair(AND, false));
      _f << "(";
      cond(_condition->getFirstChild(), _f, _logic_op, _pattern);
      _f << ")";
    }
    if(_condition->getNodeType() == DOMNode::ELEMENT_NODE &&
       !XMLString::compareString(_condition->getNodeName(), XMLString::transcode("or"))) {
      _logic_op.push(std::make_pair(OR, false));
      _f << "(";
      cond(_condition->getFirstChild(), _f, _logic_op, _pattern);
      _f << ")";
    }
    if(_condition->getNodeType() == DOMNode::ELEMENT_NODE &&
       !XMLString::compareString(_condition->getNodeName(), XMLString::transcode("test"))) {
      
      //Add logical connector if necessary
      if(_logic_op.top().second) {
	if(_logic_op.top().first == OR) {
	  _f << " || ";
	} else {
	  _f << " && ";
	}
      }
      _f << "(";
      std::string inbuffer = std::string(XMLString::transcode(_condition->getFirstChild()->getNodeValue()));
      std::istringstream input(inbuffer, std::istringstream::in);
      while(true) {
	char c;
	input.get(c);
	if(input.eof()) {
	  break;
	}
	if(c == '{') {
	  std::ostringstream vs;
	  while(true) {
	    input.get(c);
	    if(c == '}') {
	      break;
	    } else {
	      vs << c;
	    }
	  }
	  std::string substitute = lookup_arg(vs.str(), _pattern);
	  if(substitute != "") {
	    _f << substitute;
	  }
	} else {
	  _f << c;
	}
      }
      _f  << ")";
      _logic_op.top().second = true;
    }
    _condition = _condition->getNextSibling();
  }

  //Unstack
  _logic_op.pop();
  if(_logic_op.size() > 0) {
    (_logic_op.top()).second = true;
    assert(_logic_op.top().second == true);
  }
}

// void Architecture::C_condition(DOMNode *_condition, std::ofstream & _f, std::stack< int > & _logic_op, 
// 			       pattern_t *_pattern, bool b_first)
// {
//   int AND = 1;
//   int OR = 2;
//   int OUT = 3;
//   static bool b_test = false;

//   while(_condition != NULL) {
//     if(_condition->getNodeType() == DOMNode::ELEMENT_NODE &&
//        !XMLString::compareString(_condition->getNodeName(), XMLString::transcode("and"))) {
//       b_out = false;
//       if(b_test) {
// 	int curr_op = AND; //default logical operator
// 	if(_logic_op.size() > 0) {
// 	  curr_op = _logic_op.top();
// 	}
// 	if(curr_op == AND) {
// 	  _f << " && ";
// 	} else {
// 	  _f << " || ";
// 	}
//       }
//       b_test = false;

//       _f << "(";
//       _logic_op.push(AND);
//       C_condition(_condition->getFirstChild(), _f, _logic_op, _pattern);
//       _f << ")";
//       b_out = true;
//       _logic_op.pop();
//     } else if(_condition->getNodeType() == DOMNode::ELEMENT_NODE &&
//        !XMLString::compareString(_condition->getNodeName(), XMLString::transcode("or"))) {
//       b_out = false;

//       int curr_op = AND; //default logical operator
//       if(_logic_op.size() > 0) {
// 	curr_op = _logic_op.top();
//       }
//       if(curr_op == AND) {
// 	_f << " && ";
//       } else {
// 	_f << " || ";
//       }
//       b_test = false;
//       _f << "(";
//       _logic_op.push(OR);
//       C_condition(_condition->getFirstChild(), _f, _logic_op, _pattern);
//       _f << ")";
//       b_out = true;
//       _logic_op.pop();
//     } else if(_condition->getNodeType() == DOMNode::ELEMENT_NODE &&
//        !XMLString::compareString(_condition->getNodeName(), XMLString::transcode("test"))) {

//       b_test = true;
//       if(!b_first || b_out) {
// 	int curr_op = AND; //default logical operator
// 	if(_logic_op.size()) {
// 	  curr_op = _logic_op.top();
// 	}
// 	if(curr_op == AND) {
// 	  _f << " && ";
// 	} else {
// 	  _f << " || ";
// 	}
//       }
//       b_first = false;
//       b_out = false;
//       _f << "(";

//       std::string inbuffer = std::string(XMLString::transcode(_condition->getFirstChild()->getNodeValue()));
//       std::istringstream input(inbuffer, std::istringstream::in);
//       while(true) {
// 	char c;
// 	input.get(c);
// 	if(input.eof()) {
// 	  break;
// 	}
// 	if(c == '{') {
// 	  std::ostringstream vs;
// 	  while(true) {
// 	    input.get(c);
// 	    if(c == '}') {
// 	      break;
// 	    } else {
// 	      vs << c;
// 	    }
// 	  }
// 	  std::string substitute = lookup_arg(vs.str(), _pattern);
// 	  if(substitute != "") {
// 	    _f << substitute;
// 	  }
// 	} else {
// 	  _f << c;
// 	}
//       }
//       //      _f << "(" << XMLString::transcode(_condition->getFirstChild()->getNodeValue())

//       _f  << ")";
//     }
//     _condition = _condition->getNextSibling();
//   }
// }

void Architecture::start_cond_file(std::ofstream & _file)
{
  _file << "#include <unistd.h>" << std::endl
	<< "#include <sys/types.h>" << std::endl
	<< "#include <math.h>" << std::endl
	<< "#include <vector>" << std::endl
	<< "namespace lcc {" << std::endl
	<< "#ifdef __cplusplus" << std::endl
	<< "extern \"C\" {" << std::endl
	<< "#endif" << std::endl
	<< "#ifdef __GNUG__" << std::endl
	<< "#undef export" << std::endl
	<< "#define export _export" << std::endl
	<< "#define swap _swap" << std::endl
	<< "#endif //__GNUG__" << std::endl
	<< "#define typename _typename" << std::endl
	<< "#include \"c.h\"" << std::endl
	<< "#ifdef __GNUG__" << std::endl
	<< "#undef export" << std::endl
	<< "#undef swap" << std::endl
	<< "#endif //__GNUG__" << std::endl
	<< "#undef typename" << std::endl
	<< "#ifdef __cplusplus" << std::endl
	<< "}" << std::endl
	<< "#endif" << std::endl
	<< "} // namespace lcc" << std::endl
	<< std::endl
	<< "typedef bool (*condition)(std::vector< lcc::Node >);" << std::endl
	<< std::endl;
  char *tmp = get_defined_string("condition_header");
  if(tmp) {
    _file << tmp;
  }
  _file << std::endl;
}

void Architecture::end_cond_file(std::ofstream & _file)
{
  _file << "condition conditions[" << cond_index << "] = {" << std::endl;
  for(int i = 0; i < cond_index - 1; i++) {
    _file << "  __cond_" << i << "," << std::endl;
  }
  if(cond_index > 0) {
    _file << "  __cond_" << cond_index - 1 << std::endl;
  }
  _file << "};" << std::endl;
}

std::string Architecture::lookup_arg(std::string _op_name, pattern_t *_pattern)
{
  if(_pattern == NULL) {
    //case for an instruction: only src1, src2 and dest1
    if(_op_name == std::string("src1")) return std::string("(_args[0])");
    if(_op_name == std::string("src2")) return std::string("(_args[1])");
    if(_op_name == std::string("dest1")) return std::string("(_args[2])");
  } else {
    //Handle the pattern relation - index of parameter corresponds to id index in pattern
    vertex_info_map::type vim = vertex_info_property(*_pattern);
    pvertex_iter vi, vi_end;
    for(tie(vi, vi_end) = boost::vertices(*_pattern);
	vi != vi_end; ++vi) {
      if(std::string(vim[*vi].id) == _op_name) {
	//matches
	char arg[15];
	sprintf(arg, "(_args[%d])",(int)*vi);
	return std::string(arg);
      }
    }
  }
  std::cerr << "Argument: " << _op_name << " not found!" << std::endl;
  exit(-1);
  return "";
}

int Architecture::get_op_number(char *_opname)
{
  std::map< char *, int, ltstr>::iterator it = ops_map.find(_opname);
  if(it == ops_map.end()) {
    std::cerr << "Operator " << _opname << " not found in operands map (invalid LCC-IR node).\n";
    
    if(Verbose::verbosityLevel > 1) {
      it = ops_map.begin();
      while(ops_map.end() != it) {
	std::cerr << (*it).first << " : " << (*it).second << std::endl;
	it++;
      }
    }
    exit(-1);
  }
  return (*it).second;
}
